@@ -18,42 +18,21 @@
 # include <inttypes.h>
 #endif
 
+#include "base_decls.h"
 #include "buffer.h"
 #include "array.h"
 #include "chunk.h"
 #include "keyvalue.h"
-#include "fdevent.h"
-#include "sys-socket.h"
-#include "splaytree.h"
+#include "sock_addr.h"
 #include "etag.h"
 
-
-#if defined HAVE_LIBSSL && defined HAVE_OPENSSL_SSL_H
-# define USE_OPENSSL
-# include <openssl/opensslconf.h>
-#  ifndef USE_OPENSSL_KERBEROS
-#   ifndef OPENSSL_NO_KRB5
-#   define OPENSSL_NO_KRB5
-#   endif
-#  endif
-# include <openssl/ssl.h>
-# if ! defined OPENSSL_NO_TLSEXT && ! defined SSL_CTRL_SET_TLSEXT_HOSTNAME
-#  define OPENSSL_NO_TLSEXT
-# endif
-#endif
-
-#ifdef HAVE_FAM_H
-# include <fam.h>
-#endif
+struct fdevents;        /* declaration */
+struct stat_cache;      /* declaration */
 
 #ifndef O_BINARY
 # define O_BINARY 0
 #endif
 
-#ifndef O_LARGEFILE
-# define O_LARGEFILE 0
-#endif
-
 #ifndef SIZE_MAX
 # ifdef SIZE_T_MAX
 #  define SIZE_MAX SIZE_T_MAX
@@ -120,25 +99,6 @@ typedef struct {
 	char *value;
 } request_handler;
 
-typedef struct {
-	char *key;
-	char *host;
-	unsigned short port;
-	int used;
-	short factor;
-} fcgi_connections;
-
-
-typedef union {
-#ifdef HAVE_IPV6
-	struct sockaddr_in6 ipv6;
-#endif
-	struct sockaddr_in ipv4;
-#ifdef HAVE_SYS_UN_H
-	struct sockaddr_un un;
-#endif
-	struct sockaddr plain;
-} sock_addr;
 
 /* fcgi_response_header contains ... */
 #define HTTP_STATUS         BV(0)
@@ -146,6 +106,10 @@ typedef union {
 #define HTTP_CONTENT_LENGTH BV(2)
 #define HTTP_DATE           BV(3)
 #define HTTP_LOCATION       BV(4)
+#define HTTP_TRANSFER_ENCODING BV(5)
+#define HTTP_CONTENT_LOCATION  BV(6)
+#define HTTP_SET_COOKIE        BV(7)
+#define HTTP_UPGRADE           BV(8)
 
 typedef struct {
 	/** HEADER */
@@ -232,19 +196,6 @@ typedef struct {
 	buffer *content_type;
 } stat_cache_entry;
 
-typedef struct {
-	splay_tree *files; /* the nodes of the tree are stat_cache_entry's */
-
-	buffer *dir_name; /* for building the dirname from the filename */
-#ifdef HAVE_FAM_H
-	splay_tree *dirs; /* the nodes of the tree are fam_dir_entry */
-
-	FAMConnection fam;
-	int    fam_fcce_ndx;
-#endif
-	buffer *hash_key;  /* temp-store for the hash-key */
-} stat_cache;
-
 typedef struct {
 	array *mimetypes;
 
@@ -256,6 +207,7 @@ typedef struct {
 	buffer *server_tag;
 	buffer *dirlist_encoding;
 	buffer *errorfile_prefix;
+	buffer *socket_perms;
 
 	unsigned short high_precision_timestamps;
 	unsigned short max_keep_alive_requests;
@@ -267,6 +219,7 @@ typedef struct {
 	unsigned short range_requests;
 	unsigned short stream_request_body;
 	unsigned short stream_response_body;
+	unsigned short error_intercept;
 
 	/* debug */
 
@@ -275,28 +228,10 @@ typedef struct {
 	unsigned short log_request_handling;
 	unsigned short log_response_header;
 	unsigned short log_condition_handling;
-	unsigned short log_ssl_noise;
 	unsigned short log_timeouts;
 
 
 	/* server wide */
-	buffer *ssl_pemfile;
-	buffer *ssl_ca_file;
-	buffer *ssl_cipher_list;
-	buffer *ssl_dh_file;
-	buffer *ssl_ec_curve;
-	unsigned short ssl_honor_cipher_order; /* determine SSL cipher in server-preferred order, not client-order */
-	unsigned short ssl_empty_fragments; /* whether to not set SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS */
-	unsigned short ssl_use_sslv2;
-	unsigned short ssl_use_sslv3;
-	unsigned short ssl_verifyclient;
-	unsigned short ssl_verifyclient_enforce;
-	unsigned short ssl_verifyclient_depth;
-	buffer *ssl_verifyclient_username;
-	unsigned short ssl_verifyclient_export_cert;
-	unsigned short ssl_disable_client_renegotiation;
-	unsigned short ssl_read_ahead;
-
 	unsigned short use_ipv6, set_v6only; /* set_v6only is only a temporary option */
 	unsigned short defer_accept;
 	unsigned short ssl_enabled; /* only interesting for setting up listening sockets. don't use at runtime */
@@ -335,13 +270,6 @@ typedef struct {
 	buffer *bsd_accept_filter;
 #endif
 
-#ifdef USE_OPENSSL
-	SSL_CTX *ssl_ctx; /* not patched */
-	/* SNI per host: with COMP_SERVER_SOCKET, COMP_HTTP_SCHEME, COMP_HTTP_HOST */
-	EVP_PKEY *ssl_pemfile_pkey;
-	X509 *ssl_pemfile_x509;
-	STACK_OF(X509_NAME) *ssl_ca_file_cert_names;
-#endif
 } specific_config;
 
 /* the order of the items should be the same as they are processed
@@ -373,7 +301,7 @@ typedef enum {
 
 	/* actually evaluated the condition itself */
 	COND_RESULT_FALSE, /* not active */
-	COND_RESULT_TRUE, /* active */
+	COND_RESULT_TRUE   /* active */
 } cond_result_t;
 
 typedef struct {
@@ -386,7 +314,7 @@ typedef struct {
 	buffer *comp_value; /* just a pointer */
 } cond_cache_t;
 
-typedef struct {
+struct connection {
 	connection_state_t state;
 
 	/* timestamps */
@@ -448,11 +376,6 @@ typedef struct {
 
 	array  *environment; /* used to pass lighttpd internal stuff to the FastCGI/CGI apps, setenv does that */
 
-	/* response */
-	int    got_response;
-
-	int    in_joblist;
-
 	connection_type mode;
 
 	void **plugin_ctx;           /* plugin connection specific config */
@@ -461,25 +384,21 @@ typedef struct {
 	cond_cache_t *cond_cache;
 
 	buffer *server_name;
+	buffer *proto;
 
 	/* error-handler */
 	int error_handler_saved_status;
 	http_method_t error_handler_saved_method;
 
 	struct server_socket *srv_socket;   /* reference to the server-socket */
+	int (* network_write)(struct server *srv, struct connection *con, chunkqueue *cq, off_t max_bytes);
+	int (* network_read)(struct server *srv, struct connection *con, chunkqueue *cq, off_t max_bytes);
 
-#ifdef USE_OPENSSL
-	SSL *ssl;
-# ifndef OPENSSL_NO_TLSEXT
-	buffer *tlsext_server_name;
-# endif
-	unsigned int renegotiations; /* count of SSL_CB_HANDSHAKE_START */
-#endif
 	/* etag handling */
 	etag_flags_t etag_flags;
 
 	int conditional_is_valid[COMP_LAST_ELEMENT]; 
-} connection;
+};
 
 typedef struct {
 	connection **ptr;
@@ -488,19 +407,6 @@ typedef struct {
 } connections;
 
 
-#ifdef HAVE_IPV6
-typedef struct {
-	int family;
-	union {
-		struct in6_addr ipv6;
-		struct in_addr  ipv4;
-	} addr;
-	char b2[INET6_ADDRSTRLEN + 1];
-	time_t ts;
-} inet_ntop_cache_type;
-#endif
-
-
 typedef struct {
 	buffer *uri;
 	time_t mtime;
@@ -550,13 +456,7 @@ typedef struct {
 	unsigned short log_request_header_on_error;
 	unsigned short log_state_handling;
 
-	enum { STAT_CACHE_ENGINE_UNSET,
-			STAT_CACHE_ENGINE_NONE,
-			STAT_CACHE_ENGINE_SIMPLE
-#ifdef HAVE_FAM_H
-			, STAT_CACHE_ENGINE_FAM
-#endif
-	} stat_cache_engine;
+	int stat_cache_engine;
 	unsigned short enable_cores;
 	unsigned short reject_expect_100_with_417;
 	buffer *xattr_name;
@@ -567,6 +467,7 @@ typedef struct {
 	unsigned short high_precision_timestamps;
 	time_t loadts;
 	double loadavg[3];
+	buffer *syslog_facility;
 } server_config;
 
 typedef struct server_socket {
@@ -575,12 +476,9 @@ typedef struct server_socket {
 	int       fde_ndx;
 
 	unsigned short is_ssl;
+	unsigned short sidx;
 
 	buffer *srv_token;
-
-#ifdef USE_OPENSSL
-	SSL_CTX *ssl_ctx;
-#endif
 } server_socket;
 
 typedef struct {
@@ -590,7 +488,7 @@ typedef struct {
 	size_t used;
 } server_socket_array;
 
-typedef struct server {
+struct server {
 	server_socket_array srv_sockets;
 
 	/* the errorlog */
@@ -598,7 +496,7 @@ typedef struct server {
 	enum { ERRORLOG_FILE, ERRORLOG_FD, ERRORLOG_SYSLOG, ERRORLOG_PIPE } errorlog_mode;
 	buffer *errorlog_buf;
 
-	fdevents *ev, *ev_ins;
+	struct fdevents *ev;
 
 	buffer_plugin plugins;
 	void *plugin_slots;
@@ -609,8 +507,6 @@ typedef struct server {
 	int con_written;
 	int con_closed;
 
-	int ssl_is_init;
-
 	int max_fds;    /* max possible fds */
 	int cur_fds;    /* currently used fds */
 	int want_fds;   /* waiting fds */
@@ -631,9 +527,6 @@ typedef struct server {
 	buffer *cond_check_buf;
 
 	/* caches */
-#ifdef HAVE_IPV6
-	inet_ntop_cache_type inet_ntop_cache[INET_NTOP_CACHE_MAX];
-#endif
 	mtime_cache_type mtime_cache[FILE_CACHE_MAX];
 
 	array *split_vals;
@@ -662,7 +555,7 @@ typedef struct server {
 	connections *joblist;
 	connections *fdwaitqueue;
 
-	stat_cache  *stat_cache;
+	struct stat_cache *stat_cache;
 
 	/**
 	 * The status array can carry all the status information you want
@@ -679,16 +572,15 @@ typedef struct server {
 	 */
 	array *status;
 
-	fdevent_handler_t event_handler;
+	int event_handler;
 
-	int (* network_backend_write)(struct server *srv, connection *con, int fd, chunkqueue *cq, off_t max_bytes);
-#ifdef USE_OPENSSL
-	int (* network_ssl_backend_write)(struct server *srv, connection *con, SSL *ssl, chunkqueue *cq, off_t max_bytes);
-#endif
+	int (* network_backend_write)(struct server *srv, int fd, chunkqueue *cq, off_t max_bytes);
+	handler_t (* request_env)(struct server *srv, connection *con);
 
 	uid_t uid;
 	gid_t gid;
-} server;
+	pid_t pid;
+};
 
 
 #endif
