@@ -4,9 +4,7 @@
 
 #include "settings.h"
 
-#include <stdlib.h>
 #include <sys/types.h>
-#include <stdio.h>
 #include <time.h>
 
 #if defined HAVE_STDINT_H
@@ -101,15 +99,13 @@ void li_utostrn(char *buf, size_t buf_len, uintmax_t val);
 /* buf must be (at least) 2*s_len + 1 big. uses lower-case hex letters. */
 void li_tohex(char *buf, size_t buf_len, const char *s, size_t s_len);
 
-char * buffer_search_string_len(buffer *b, const char *needle, size_t len);
-
 /* NULL buffer or empty buffer (used == 0);
  * unset "string" (buffer) config options are initialized to used == 0,
  * while setting an empty string leads to used == 1
  */
-int buffer_is_empty(const buffer *b);
+static inline int buffer_is_empty(const buffer *b);
 /* NULL buffer, empty buffer (used == 0) or empty string (used == 1) */
-int buffer_string_is_empty(const buffer *b);
+static inline int buffer_string_is_empty(const buffer *b);
 
 int buffer_is_equal(const buffer *a, const buffer *b);
 int buffer_is_equal_right_len(const buffer *a, const buffer *b, size_t len);
@@ -117,6 +113,8 @@ int buffer_is_equal_string(const buffer *a, const char *s, size_t b_len);
 int buffer_is_equal_caseless_string(const buffer *a, const char *s, size_t b_len);
 int buffer_caseless_compare(const char *a, size_t a_len, const char *b, size_t b_len);
 
+void buffer_substr_replace (buffer *b, size_t offset, size_t len, const buffer *replace);
+
 typedef enum {
 	ENCODING_REL_URI, /* for coding a rel-uri (/with space/and%percent) nicely as part of a href */
 	ENCODING_REL_URI_PART, /* same as ENC_REL_URL plus coding / too as %2F */
@@ -165,13 +163,19 @@ static inline void buffer_append_slash(buffer *b); /* append '/' no non-empty st
 #define CONST_BUF_LEN(x) ((x) ? (x)->ptr : NULL), buffer_string_length(x)
 
 
-void print_backtrace(FILE *file);
 void log_failed_assert(const char *filename, unsigned int line, const char *msg) LI_NORETURN;
 #define force_assert(x) do { if (!(x)) log_failed_assert(__FILE__, __LINE__, "assertion failed: " #x); } while(0)
 #define SEGFAULT() log_failed_assert(__FILE__, __LINE__, "aborted");
 
 /* inline implementations */
 
+static inline int buffer_is_empty(const buffer *b) {
+	return NULL == b || 0 == b->used;
+}
+static inline int buffer_string_is_empty(const buffer *b) {
+	return NULL == b || b->used < 2;
+}
+
 static inline size_t buffer_string_length(const buffer *b) {
 	return NULL != b && 0 != b->used ? b->used - 1 : 0;
 }
