@@ -1,17 +1,12 @@
 #include "first.h"
 
+#include "sys-socket.h"
 #include "base.h"
 #include "connections.h"
-#include "joblist.h"
 #include "log.h"
 
 #include <errno.h>
 
-#ifdef USE_OPENSSL
-# include <openssl/ssl.h>
-# include <openssl/err.h>
-#endif
-
 const char *connection_get_state(connection_state_t state) {
 	switch (state) {
 	case CON_STATE_CONNECT: return "connect";
@@ -54,269 +49,6 @@ int connection_set_state(server *srv, connection *con, connection_state_t state)
 	return 0;
 }
 
-#if 0
-static void dump_packet(const unsigned char *data, size_t len) {
-	size_t i, j;
-
-	if (len == 0) return;
-
-	for (i = 0; i < len; i++) {
-		if (i % 16 == 0) fprintf(stderr, "  ");
-
-		fprintf(stderr, "%02x ", data[i]);
-
-		if ((i + 1) % 16 == 0) {
-			fprintf(stderr, "  ");
-			for (j = 0; j <= i % 16; j++) {
-				unsigned char c;
-
-				if (i-15+j >= len) break;
-
-				c = data[i-15+j];
-
-				fprintf(stderr, "%c", c > 32 && c < 128 ? c : '.');
-			}
-
-			fprintf(stderr, "\n");
-		}
-	}
-
-	if (len % 16 != 0) {
-		for (j = i % 16; j < 16; j++) {
-			fprintf(stderr, "   ");
-		}
-
-		fprintf(stderr, "  ");
-		for (j = i & ~0xf; j < len; j++) {
-			unsigned char c;
-
-			c = data[j];
-			fprintf(stderr, "%c", c > 32 && c < 128 ? c : '.');
-		}
-		fprintf(stderr, "\n");
-	}
-}
-#endif
-
-static int connection_handle_read_ssl(server *srv, connection *con) {
-#ifdef USE_OPENSSL
-	int r, ssl_err, len;
-	char *mem = NULL;
-	size_t mem_len = 0;
-
-	if (!con->srv_socket->is_ssl) return -1;
-
-	ERR_clear_error();
-	do {
-		chunkqueue_get_memory(con->read_queue, &mem, &mem_len, 0, SSL_pending(con->ssl));
-#if 0
-		/* overwrite everything with 0 */
-		memset(mem, 0, mem_len);
-#endif
-
-		len = SSL_read(con->ssl, mem, mem_len);
-		if (len > 0) {
-			chunkqueue_use_memory(con->read_queue, len);
-			con->bytes_read += len;
-		} else {
-			chunkqueue_use_memory(con->read_queue, 0);
-		}
-
-		if (con->renegotiations > 1 && con->conf.ssl_disable_client_renegotiation) {
-			log_error_write(srv, __FILE__, __LINE__, "s", "SSL: renegotiation initiated by client, killing connection");
-			connection_set_state(srv, con, CON_STATE_ERROR);
-			return -1;
-		}
-	} while (len > 0 && (con->conf.ssl_read_ahead || SSL_pending(con->ssl) > 0));
-
-	if (len < 0) {
-		int oerrno = errno;
-		switch ((r = SSL_get_error(con->ssl, len))) {
-		case SSL_ERROR_WANT_WRITE:
-			con->is_writable = -1;
-		case SSL_ERROR_WANT_READ:
-			con->is_readable = 0;
-
-			/* the manual says we have to call SSL_read with the same arguments next time.
-			 * we ignore this restriction; no one has complained about it in 1.5 yet, so it probably works anyway.
-			 */
-
-			return 0;
-		case SSL_ERROR_SYSCALL:
-			/**
-			 * man SSL_get_error()
-			 *
-			 * SSL_ERROR_SYSCALL
-			 *   Some I/O error occurred.  The OpenSSL error queue may contain more
-			 *   information on the error.  If the error queue is empty (i.e.
-			 *   ERR_get_error() returns 0), ret can be used to find out more about
-			 *   the error: If ret == 0, an EOF was observed that violates the
-			 *   protocol.  If ret == -1, the underlying BIO reported an I/O error
-			 *   (for socket I/O on Unix systems, consult errno for details).
-			 *
-			 */
-			while((ssl_err = ERR_get_error())) {
-				/* get all errors from the error-queue */
-				log_error_write(srv, __FILE__, __LINE__, "sds", "SSL:",
-						r, ERR_error_string(ssl_err, NULL));
-			}
-
-			switch(oerrno) {
-			default:
-				log_error_write(srv, __FILE__, __LINE__, "sddds", "SSL:",
-						len, r, oerrno,
-						strerror(oerrno));
-				break;
-			}
-
-			break;
-		case SSL_ERROR_ZERO_RETURN:
-			/* clean shutdown on the remote side */
-
-			if (r == 0) {
-				/* FIXME: later */
-			}
-
-			/* fall thourgh */
-		default:
-			while((ssl_err = ERR_get_error())) {
-				switch (ERR_GET_REASON(ssl_err)) {
-				case SSL_R_SSL_HANDSHAKE_FAILURE:
-			      #ifdef SSL_R_TLSV1_ALERT_UNKNOWN_CA
-				case SSL_R_TLSV1_ALERT_UNKNOWN_CA:
-			      #endif
-			      #ifdef SSL_R_SSLV3_ALERT_CERTIFICATE_UNKNOWN
-				case SSL_R_SSLV3_ALERT_CERTIFICATE_UNKNOWN:
-			      #endif
-			      #ifdef SSL_R_SSLV3_ALERT_BAD_CERTIFICATE
-				case SSL_R_SSLV3_ALERT_BAD_CERTIFICATE:
-			      #endif
-					if (!con->conf.log_ssl_noise) continue;
-					break;
-				default:
-					break;
-				}
-				/* get all errors from the error-queue */
-				log_error_write(srv, __FILE__, __LINE__, "sds", "SSL:",
-				                r, ERR_error_string(ssl_err, NULL));
-			}
-			break;
-		}
-
-		connection_set_state(srv, con, CON_STATE_ERROR);
-
-		return -1;
-	} else if (len == 0) {
-		con->is_readable = 0;
-		/* the other end close the connection -> KEEP-ALIVE */
-
-		return -2;
-	} else {
-		return 0;
-	}
-#else
-	UNUSED(srv);
-	UNUSED(con);
-	return -1;
-#endif
-}
-
-/* 0: everything ok, -1: error, -2: con closed */
-int connection_handle_read(server *srv, connection *con) {
-	int len;
-	char *mem = NULL;
-	size_t mem_len = 0;
-	int toread;
-
-	if (con->srv_socket->is_ssl) {
-		return connection_handle_read_ssl(srv, con);
-	}
-
-	/* default size for chunks is 4kb; only use bigger chunks if FIONREAD tells
-	 *  us more than 4kb is available
-	 * if FIONREAD doesn't signal a big chunk we fill the previous buffer
-	 *  if it has >= 1kb free
-	 */
-#if defined(__WIN32)
-	chunkqueue_get_memory(con->read_queue, &mem, &mem_len, 0, 4096);
-
-	len = recv(con->fd, mem, mem_len, 0);
-#else /* __WIN32 */
-	if (ioctl(con->fd, FIONREAD, &toread) || toread <= 4*1024) {
-		toread = 4096;
-	}
-	else if (toread > MAX_READ_LIMIT) {
-		toread = MAX_READ_LIMIT;
-	}
-	chunkqueue_get_memory(con->read_queue, &mem, &mem_len, 0, toread);
-
-	len = read(con->fd, mem, mem_len);
-#endif /* __WIN32 */
-
-	chunkqueue_use_memory(con->read_queue, len > 0 ? len : 0);
-
-	if (len < 0) {
-		con->is_readable = 0;
-
-#if defined(__WIN32)
-		{
-			int lastError = WSAGetLastError();
-			switch (lastError) {
-			case EAGAIN:
-				return 0;
-			case EINTR:
-				/* we have been interrupted before we could read */
-				con->is_readable = 1;
-				return 0;
-			case ECONNRESET:
-				/* suppress logging for this error, expected for keep-alive */
-				break;
-			default:
-				log_error_write(srv, __FILE__, __LINE__, "sd", "connection closed - recv failed: ", lastError);
-				break;
-			}
-		}
-#else /* __WIN32 */
-		switch (errno) {
-		case EAGAIN:
-			return 0;
-		case EINTR:
-			/* we have been interrupted before we could read */
-			con->is_readable = 1;
-			return 0;
-		case ECONNRESET:
-			/* suppress logging for this error, expected for keep-alive */
-			break;
-		default:
-			log_error_write(srv, __FILE__, __LINE__, "ssd", "connection closed - read failed: ", strerror(errno), errno);
-			break;
-		}
-#endif /* __WIN32 */
-
-		connection_set_state(srv, con, CON_STATE_ERROR);
-
-		return -1;
-	} else if (len == 0) {
-		con->is_readable = 0;
-		/* the other end close the connection -> KEEP-ALIVE */
-
-		/* pipelining */
-
-		return -2;
-	} else if (len != (ssize_t) mem_len) {
-		/* we got less then expected, wait for the next fd-event */
-
-		con->is_readable = 0;
-	}
-
-	con->bytes_read += len;
-#if 0
-	dump_packet(b->ptr, len);
-#endif
-
-	return 0;
-}
-
 static int connection_handle_read_post_cq_compact(chunkqueue *cq) {
     /* combine first mem chunk with next non-empty mem chunk
      * (loop if next chunk is empty) */
@@ -393,6 +125,7 @@ static handler_t connection_handle_read_post_chunked(server *srv, connection *co
         while (0 == te_chunked) {
             char *p;
             chunk *c = cq->first;
+            if (NULL == c) break;
             force_assert(c->type == MEM_CHUNK);
             p = strchr(c->mem->ptr+c->offset, '\n');
             if (NULL != p) { /* found HTTP chunked header line */
@@ -532,6 +265,146 @@ static handler_t connection_handle_read_post_chunked(server *srv, connection *co
     return HANDLER_GO_ON;
 }
 
+static handler_t connection_handle_read_body_unknown(server *srv, connection *con, chunkqueue *cq, chunkqueue *dst_cq) {
+    /* con->conf.max_request_size is in kBytes */
+    const off_t max_request_size = (off_t)con->conf.max_request_size << 10;
+    chunkqueue_append_chunkqueue(dst_cq, cq);
+    if (0 != max_request_size && dst_cq->bytes_in > max_request_size) {
+        log_error_write(srv, __FILE__, __LINE__, "sos",
+                        "request-size too long:", dst_cq->bytes_in, "-> 413");
+        /* 413 Payload Too Large */
+        return connection_handle_read_post_error(srv, con, 413);
+    }
+    return HANDLER_GO_ON;
+}
+
+static off_t connection_write_throttle(server *srv, connection *con, off_t max_bytes) {
+	UNUSED(srv);
+	if (con->conf.global_kbytes_per_second) {
+		off_t limit = con->conf.global_kbytes_per_second * 1024 - *(con->conf.global_bytes_per_second_cnt_ptr);
+		if (limit <= 0) {
+			/* we reached the global traffic limit */
+			con->traffic_limit_reached = 1;
+
+			return 0;
+		} else {
+			if (max_bytes > limit) max_bytes = limit;
+		}
+	}
+
+	if (con->conf.kbytes_per_second) {
+		off_t limit = con->conf.kbytes_per_second * 1024 - con->bytes_written_cur_second;
+		if (limit <= 0) {
+			/* we reached the traffic limit */
+			con->traffic_limit_reached = 1;
+
+			return 0;
+		} else {
+			if (max_bytes > limit) max_bytes = limit;
+		}
+	}
+
+	return max_bytes;
+}
+
+int connection_write_chunkqueue(server *srv, connection *con, chunkqueue *cq, off_t max_bytes) {
+	int ret = -1;
+	off_t written = 0;
+      #ifdef TCP_CORK
+	int corked = 0;
+      #endif
+
+	max_bytes = connection_write_throttle(srv, con, max_bytes);
+	if (0 == max_bytes) return 1;
+
+	written = cq->bytes_out;
+
+      #ifdef TCP_CORK
+	/* Linux: put a cork into socket as we want to combine write() calls
+	 * but only if we really have multiple chunks including non-MEM_CHUNK,
+	 * and only if TCP socket
+	 */
+	if (cq->first && cq->first->next) {
+		const int sa_family = sock_addr_get_family(&con->srv_socket->addr);
+		if (sa_family == AF_INET || sa_family == AF_INET6) {
+			chunk *c = cq->first;
+			while (c->type == MEM_CHUNK && NULL != (c = c->next)) ;
+			if (NULL != c) {
+				corked = 1;
+				(void)setsockopt(con->fd, IPPROTO_TCP, TCP_CORK, &corked, sizeof(corked));
+			}
+		}
+	}
+      #endif
+
+	ret = con->network_write(srv, con, cq, max_bytes);
+	if (ret >= 0) {
+		chunkqueue_remove_finished_chunks(cq);
+		ret = chunkqueue_is_empty(cq) ? 0 : 1;
+	}
+
+      #ifdef TCP_CORK
+	if (corked) {
+		corked = 0;
+		(void)setsockopt(con->fd, IPPROTO_TCP, TCP_CORK, &corked, sizeof(corked));
+	}
+      #endif
+
+	written = cq->bytes_out - written;
+	con->bytes_written += written;
+	con->bytes_written_cur_second += written;
+	*(con->conf.global_bytes_per_second_cnt_ptr) += written;
+
+	return ret;
+}
+
+static int connection_write_100_continue(server *srv, connection *con) {
+	/* Make best effort to send all or none of "HTTP/1.1 100 Continue" */
+	/* (Note: also choosing not to update con->write_request_ts
+	 *  which differs from connections.c:connection_handle_write()) */
+	static const char http_100_continue[] = "HTTP/1.1 100 Continue\r\n\r\n";
+	chunkqueue *cq;
+	off_t written;
+	int rc;
+
+	off_t max_bytes =
+	  connection_write_throttle(srv, con, sizeof(http_100_continue)-1);
+	if (max_bytes < (off_t)sizeof(http_100_continue)-1) {
+		return 1; /* success; skip sending if throttled to partial */
+	}
+
+	cq = con->write_queue;
+	written = cq->bytes_out;
+
+	chunkqueue_append_mem(cq,http_100_continue,sizeof(http_100_continue)-1);
+	rc = con->network_write(srv, con, cq, sizeof(http_100_continue)-1);
+
+	written = cq->bytes_out - written;
+	con->bytes_written += written;
+	con->bytes_written_cur_second += written;
+	*(con->conf.global_bytes_per_second_cnt_ptr) += written;
+
+	if (rc < 0) {
+		connection_set_state(srv, con, CON_STATE_ERROR);
+		return 0; /* error */
+	}
+
+	if (written == sizeof(http_100_continue)-1) {
+		chunkqueue_remove_finished_chunks(cq);
+	} else if (0 == written) {
+		/* skip sending 100 Continue if send would block */
+		chunkqueue_mark_written(cq, sizeof(http_100_continue)-1);
+		con->is_writable = 0;
+	}
+	/* else partial write (unlikely), which can cause corrupt
+	 * response if response is later cleared, e.g. sending errdoc.
+	 * However, situation of partial write can occur here only on
+	 * keep-alive request where client has sent pipelined request,
+	 * and more than 0 chars were written, but fewer than 25 chars */
+
+	return 1; /* success; sent all or none of "HTTP/1.1 100 Continue" */
+}
+
 handler_t connection_handle_read_post_state(server *srv, connection *con) {
 	chunkqueue *cq = con->read_queue;
 	chunkqueue *dst_cq = con->request_content_queue;
@@ -541,8 +414,9 @@ handler_t connection_handle_read_post_state(server *srv, connection *con) {
 	if (con->is_readable) {
 		con->read_idle_ts = srv->cur_ts;
 
-		switch(connection_handle_read(srv, con)) {
+		switch(con->network_read(srv, con, con->read_queue, MAX_READ_LIMIT)) {
 		case -1:
+			connection_set_state(srv, con, CON_STATE_ERROR);
 			return HANDLER_ERROR;
 		case -2:
 			is_closed = 1;
@@ -554,8 +428,25 @@ handler_t connection_handle_read_post_state(server *srv, connection *con) {
 
 	chunkqueue_remove_finished_chunks(cq);
 
-	if (-1 == con->request.content_length) { /*(Transfer-Encoding: chunked)*/
-		handler_t rc = connection_handle_read_post_chunked(srv, con, cq, dst_cq);
+	/* Check for Expect: 100-continue in request headers
+	 * if no request body received yet */
+	if (chunkqueue_is_empty(cq) && 0 == dst_cq->bytes_in
+	    && con->request.http_version != HTTP_VERSION_1_0
+	    && chunkqueue_is_empty(con->write_queue) && con->is_writable) {
+		data_string *ds = (data_string *)array_get_element(con->request.headers, "Expect");
+		if (NULL != ds && 0 == buffer_caseless_compare(CONST_BUF_LEN(ds->value), CONST_STR_LEN("100-continue"))) {
+			buffer_reset(ds->value); /* unset value in request headers */
+			if (!connection_write_100_continue(srv, con)) {
+				return HANDLER_ERROR;
+			}
+		}
+	}
+
+	if (con->request.content_length < 0) {
+		/*(-1: Transfer-Encoding: chunked, -2: unspecified length)*/
+		handler_t rc = (-1 == con->request.content_length)
+                  ? connection_handle_read_post_chunked(srv, con, cq, dst_cq)
+                  : connection_handle_read_body_unknown(srv, con, cq, dst_cq);
 		if (HANDLER_GO_ON != rc) return rc;
 	}
 	else if (con->request.content_length <= 64*1024) {
@@ -597,7 +488,6 @@ void connection_response_reset(server *srv, connection *con) {
 	con->is_writable = 1;
 	con->file_finished = 0;
 	con->file_started = 0;
-	con->got_response = 0;
 	con->parsed_response = 0;
 	con->response.keep_alive = 0;
 	con->response.content_length = -1;
