@@ -16,9 +16,9 @@ int connection_set_state(server *srv, connection *con, connection_state_t state)
 const char * connection_get_state(connection_state_t state);
 const char * connection_get_short_state(connection_state_t state);
 int connection_state_machine(server *srv, connection *con);
-int connection_handle_read(server *srv, connection *con);
 handler_t connection_handle_read_post_state(server *srv, connection *con);
 handler_t connection_handle_read_post_error(server *srv, connection *con, int http_status);
+int connection_write_chunkqueue(server *srv, connection *con, chunkqueue *c, off_t max_bytes);
 void connection_response_reset(server *srv, connection *con);
 
 #endif
