@@ -3,10 +3,12 @@
 #include "base.h"
 #include "array.h"
 #include "buffer.h"
+#include "fdevent.h"
 #include "log.h"
 #include "etag.h"
 #include "http_chunk.h"
 #include "response.h"
+#include "sock_addr.h"
 #include "stat_cache.h"
 
 #include <string.h>
@@ -14,66 +16,9 @@
 
 #include <time.h>
 
-/*
- * This was 'borrowed' from tcpdump.
- *
- *
- * This is fun.
- *
- * In older BSD systems, socket addresses were fixed-length, and
- * "sizeof (struct sockaddr)" gave the size of the structure.
- * All addresses fit within a "struct sockaddr".
- *
- * In newer BSD systems, the socket address is variable-length, and
- * there's an "sa_len" field giving the length of the structure;
- * this allows socket addresses to be longer than 2 bytes of family
- * and 14 bytes of data.
- *
- * Some commercial UNIXes use the old BSD scheme, some use the RFC 2553
- * variant of the old BSD scheme (with "struct sockaddr_storage" rather
- * than "struct sockaddr"), and some use the new BSD scheme.
- *
- * Some versions of GNU libc use neither scheme, but has an "SA_LEN()"
- * macro that determines the size based on the address family.  Other
- * versions don't have "SA_LEN()" (as it was in drafts of RFC 2553
- * but not in the final version).  On the latter systems, we explicitly
- * check the AF_ type to determine the length; we assume that on
- * all those systems we have "struct sockaddr_storage".
- */
-
-#ifdef HAVE_IPV6
-# ifndef SA_LEN
-#  ifdef HAVE_SOCKADDR_SA_LEN
-#   define SA_LEN(addr)   ((addr)->sa_len)
-#  else /* HAVE_SOCKADDR_SA_LEN */
-#   ifdef HAVE_STRUCT_SOCKADDR_STORAGE
-static size_t get_sa_len(const struct sockaddr *addr) {
-	switch (addr->sa_family) {
-
-#    ifdef AF_INET
-	case AF_INET:
-		return (sizeof (struct sockaddr_in));
-#    endif
-
-#    ifdef AF_INET6
-	case AF_INET6:
-		return (sizeof (struct sockaddr_in6));
-#    endif
-
-	default:
-		return (sizeof (struct sockaddr));
-
-	}
-}
-#    define SA_LEN(addr)   (get_sa_len(addr))
-#   else /* HAVE_SOCKADDR_STORAGE */
-#    define SA_LEN(addr)   (sizeof (struct sockaddr))
-#   endif /* HAVE_SOCKADDR_STORAGE */
-#  endif /* HAVE_SOCKADDR_SA_LEN */
-# endif /* SA_LEN */
-#endif
-
-
+#include "sys-strings.h"
+#include "sys-socket.h"
+#include <unistd.h>
 
 
 int response_header_insert(server *srv, connection *con, const char *key, size_t keylen, const char *value, size_t vallen) {
@@ -98,8 +43,8 @@ int response_header_overwrite(server *srv, connection *con, const char *key, siz
 	UNUSED(srv);
 
 	/* if there already is a key by this name overwrite the value */
-	if (NULL != (ds = (data_string *)array_get_element(con->response.headers, key))) {
-		buffer_copy_string(ds->value, value);
+	if (NULL != (ds = (data_string *)array_get_element_klen(con->response.headers, key, keylen))) {
+		buffer_copy_string_len(ds->value, value, vallen);
 
 		return 0;
 	}
@@ -113,7 +58,7 @@ int response_header_append(server *srv, connection *con, const char *key, size_t
 	UNUSED(srv);
 
 	/* if there already is a key by this name append the value */
-	if (NULL != (ds = (data_string *)array_get_element(con->response.headers, key))) {
+	if (NULL != (ds = (data_string *)array_get_element_klen(con->response.headers, key, keylen))) {
 		buffer_append_string_len(ds->value, CONST_STR_LEN(", "));
 		buffer_append_string_len(ds->value, value, vallen);
 		return 0;
@@ -133,10 +78,6 @@ int http_response_redirect_to_directory(server *srv, connection *con) {
 		buffer_append_string_buffer(o, con->uri.authority);
 	} else {
 		/* get the name of the currently connected socket */
-		struct hostent *he;
-#ifdef HAVE_IPV6
-		char hbuf[256];
-#endif
 		sock_addr our_addr;
 		socklen_t our_addr_len;
 
@@ -153,51 +94,12 @@ int http_response_redirect_to_directory(server *srv, connection *con) {
 			return 0;
 		}
 
-
 		/* Lookup name: secondly try to get hostname for bind address */
-		switch(our_addr.plain.sa_family) {
-#ifdef HAVE_IPV6
-		case AF_INET6:
-			if (0 != getnameinfo((const struct sockaddr *)(&our_addr.ipv6),
-					     SA_LEN((const struct sockaddr *)&our_addr.ipv6),
-					     hbuf, sizeof(hbuf), NULL, 0, 0)) {
-
-				char dst[INET6_ADDRSTRLEN];
-
-				log_error_write(srv, __FILE__, __LINE__,
-						"SSS", "NOTICE: getnameinfo failed: ",
-						strerror(errno), ", using ip-address instead");
-
-				buffer_append_string_len(o, CONST_STR_LEN("["));
-				buffer_append_string(o,
-						     inet_ntop(AF_INET6, (char *)&our_addr.ipv6.sin6_addr,
-							       dst, sizeof(dst)));
-				buffer_append_string_len(o, CONST_STR_LEN("]"));
-			} else {
-				buffer_append_string(o, hbuf);
-			}
-			break;
-#endif
-		case AF_INET:
-			if (NULL == (he = gethostbyaddr((char *)&our_addr.ipv4.sin_addr, sizeof(struct in_addr), AF_INET))) {
-				log_error_write(srv, __FILE__, __LINE__,
-						"SdS", "NOTICE: gethostbyaddr failed: ",
-						h_errno, ", using ip-address instead");
-
-				buffer_append_string(o, inet_ntoa(our_addr.ipv4.sin_addr));
-			} else {
-				buffer_append_string(o, he->h_name);
-			}
-			break;
-		default:
-			log_error_write(srv, __FILE__, __LINE__,
-					"S", "ERROR: unsupported address-type");
-
+		if (0 != sock_addr_nameinfo_append_buffer(srv, o, &our_addr)) {
+			con->http_status = 500;
 			buffer_free(o);
 			return -1;
-		}
-
-		{
+		} else {
 			unsigned short default_port = 80;
 			if (buffer_is_equal_caseless_string(con->uri.scheme, CONST_STR_LEN("https"))) {
 				default_port = 443;
@@ -226,24 +128,21 @@ int http_response_redirect_to_directory(server *srv, connection *con) {
 }
 
 buffer * strftime_cache_get(server *srv, time_t last_mod) {
+	static int i;
 	struct tm *tm;
-	size_t i;
 
-	for (i = 0; i < FILE_CACHE_MAX; i++) {
-		/* found cache-entry */
-		if (srv->mtime_cache[i].mtime == last_mod) return srv->mtime_cache[i].str;
-
-		/* found empty slot */
-		if (srv->mtime_cache[i].mtime == 0) break;
+	for (int j = 0; j < FILE_CACHE_MAX; ++j) {
+		if (srv->mtime_cache[j].mtime == last_mod)
+			return srv->mtime_cache[j].str; /* found cache-entry */
 	}
 
-	if (i == FILE_CACHE_MAX) {
+	if (++i == FILE_CACHE_MAX) {
 		i = 0;
 	}
 
 	srv->mtime_cache[i].mtime = last_mod;
-	buffer_string_prepare_copy(srv->mtime_cache[i].str, 1023);
 	tm = gmtime(&(srv->mtime_cache[i].mtime));
+	buffer_string_set_length(srv->mtime_cache[i].str, 0);
 	buffer_append_strftime(srv->mtime_cache[i].str, "%a, %d %b %Y %H:%M:%S GMT", tm);
 
 	return srv->mtime_cache[i].str;
@@ -670,12 +569,11 @@ void http_response_send_file (server *srv, connection *con, buffer *path) {
 	}
 }
 
-void http_response_xsendfile (server *srv, connection *con, buffer *path, const array *xdocroot) {
+
+static void http_response_xsendfile (server *srv, connection *con, buffer *path, const array *xdocroot) {
 	const int status = con->http_status;
 	int valid = 1;
 
-	con->file_started = 1;
-
 	/* reset Content-Length, if set by backend
 	 * Content-Length might later be set to size of X-Sendfile static file,
 	 * determined by open(), fstat() to reduces race conditions if the file
@@ -727,6 +625,143 @@ void http_response_xsendfile (server *srv, connection *con, buffer *path, const
 	}
 }
 
+
+static void http_response_xsendfile2(server *srv, connection *con, const buffer *value, const array *xdocroot) {
+    const char *pos = value->ptr;
+    buffer *b = srv->tmp_buf;
+    const int status = con->http_status;
+
+    /* reset Content-Length, if set by backend */
+    if (con->parsed_response & HTTP_CONTENT_LENGTH) {
+        data_string *ds = (data_string *)
+          array_get_element(con->response.headers, "Content-Length");
+        if (ds) buffer_reset(ds->value);
+        con->parsed_response &= ~HTTP_CONTENT_LENGTH;
+        con->response.content_length = -1;
+    }
+
+    while (*pos) {
+        const char *filename, *range;
+        stat_cache_entry *sce;
+        off_t begin_range, end_range, range_len;
+
+        while (' ' == *pos) pos++;
+        if (!*pos) break;
+
+        filename = pos;
+        if (NULL == (range = strchr(pos, ' '))) {
+            /* missing range */
+            log_error_write(srv, __FILE__, __LINE__, "ss",
+                            "Couldn't find range after filename:", filename);
+            con->http_status = 502;
+            break;
+        }
+        buffer_copy_string_len(b, filename, range - filename);
+
+        /* find end of range */
+        for (pos = ++range; *pos && *pos != ' ' && *pos != ','; pos++) ;
+
+        buffer_urldecode_path(b);
+        buffer_path_simplify(b, b);
+        if (con->conf.force_lowercase_filenames) {
+            buffer_to_lower(b);
+        }
+        if (xdocroot->used) {
+            size_t i, xlen = buffer_string_length(b);
+            for (i = 0; i < xdocroot->used; ++i) {
+                data_string *ds = (data_string *)xdocroot->data[i];
+                size_t dlen = buffer_string_length(ds->value);
+                if (dlen <= xlen
+                    && (!con->conf.force_lowercase_filenames
+                    ? 0 == memcmp(b->ptr, ds->value->ptr, dlen)
+                    : 0 == strncasecmp(b->ptr, ds->value->ptr, dlen))) {
+                    break;
+                }
+            }
+            if (i == xdocroot->used) {
+                log_error_write(srv, __FILE__, __LINE__, "SBs",
+                                "X-Sendfile2 (", b,
+                                ") not under configured x-sendfile-docroot(s)");
+                con->http_status = 403;
+                break;
+            }
+        }
+
+        if (HANDLER_ERROR == stat_cache_get_entry(srv, con, b, &sce)) {
+            log_error_write(srv, __FILE__, __LINE__, "sb", "send-file error: "
+                            "couldn't get stat_cache entry for X-Sendfile2:",
+                            b);
+            con->http_status = 404;
+            break;
+        } else if (!S_ISREG(sce->st.st_mode)) {
+            log_error_write(srv, __FILE__, __LINE__, "sb",
+                            "send-file error: wrong filetype for X-Sendfile2:",
+                            b);
+            con->http_status = 502;
+            break;
+        }
+        /* found the file */
+
+        /* parse range */
+        end_range = sce->st.st_size - 1;
+        {
+            char *rpos = NULL;
+            errno = 0;
+            begin_range = strtoll(range, &rpos, 10);
+            if (errno != 0 || begin_range < 0 || rpos == range)
+                goto range_failed;
+            if ('-' != *rpos++) goto range_failed;
+            if (rpos != pos) {
+                range = rpos;
+                end_range = strtoll(range, &rpos, 10);
+                if (errno != 0 || end_range < 0 || rpos == range)
+                    goto range_failed;
+            }
+            if (rpos != pos) goto range_failed;
+
+            goto range_success;
+
+range_failed:
+            log_error_write(srv, __FILE__, __LINE__, "ss",
+                            "Couldn't decode range after filename:", filename);
+            con->http_status = 502;
+            break;
+
+range_success: ;
+        }
+
+        /* no parameters accepted */
+
+        while (*pos == ' ') pos++;
+        if (*pos != '\0' && *pos != ',') {
+            con->http_status = 502;
+            break;
+        }
+
+        range_len = end_range - begin_range + 1;
+        if (range_len < 0) {
+            con->http_status = 502;
+            break;
+        }
+        if (range_len != 0) {
+            if (0 != http_chunk_append_file_range(srv, con, b,
+                                                  begin_range, range_len)) {
+                con->http_status = 502;
+                break;
+            }
+        }
+
+        if (*pos == ',') pos++;
+    }
+
+    if (con->http_status >= 400 && status < 300) {
+        con->mode = DIRECT;
+    } else if (0 != status && 200 != status) {
+        con->http_status = status;
+    }
+}
+
+
 void http_response_backend_error (server *srv, connection *con) {
 	UNUSED(srv);
 	if (con->file_started) {
@@ -762,21 +797,533 @@ void http_response_backend_done (server *srv, connection *con) {
 }
 
 
+void http_response_upgrade_read_body_unknown(server *srv, connection *con) {
+    /* act as transparent proxy */
+    UNUSED(srv);
+    if (!(con->conf.stream_request_body & FDEVENT_STREAM_REQUEST))
+        con->conf.stream_request_body |=
+          (FDEVENT_STREAM_REQUEST_BUFMIN | FDEVENT_STREAM_REQUEST);
+    if (!(con->conf.stream_response_body & FDEVENT_STREAM_RESPONSE))
+        con->conf.stream_response_body |=
+          (FDEVENT_STREAM_RESPONSE_BUFMIN | FDEVENT_STREAM_RESPONSE);
+    con->conf.stream_request_body |= FDEVENT_STREAM_REQUEST_POLLIN;
+    con->request.content_length = -2;
+    con->keep_alive = 0;
+}
+
+
+static handler_t http_response_process_local_redir(server *srv, connection *con, size_t blen) {
+    /* [RFC3875] The Common Gateway Interface (CGI) Version 1.1
+     * [RFC3875] 6.2.2 Local Redirect Response
+     *
+     *    The CGI script can return a URI path and query-string
+     *    ('local-pathquery') for a local resource in a Location header field.
+     *    This indicates to the server that it should reprocess the request
+     *    using the path specified.
+     *
+     *      local-redir-response = local-Location NL
+     *
+     *    The script MUST NOT return any other header fields or a message-body,
+     *    and the server MUST generate the response that it would have produced
+     *    in response to a request containing the URL
+     *
+     *      scheme "://" server-name ":" server-port local-pathquery
+     *
+     * (Might not have begun to receive body yet, but do skip local-redir
+     *  if we already have started receiving a response body (blen > 0))
+     * (Also, while not required by the RFC, do not send local-redir back
+     *  to same URL, since CGI should have handled it internally if it
+     *  really wanted to do that internally)
+     */
+
+    /* con->http_status >= 300 && con->http_status < 400) */
+    size_t ulen = buffer_string_length(con->uri.path);
+    data_string *ds = (data_string *)
+      array_get_element(con->response.headers, "Location");
+    if (NULL != ds
+        && ds->value->ptr[0] == '/'
+        && (0 != strncmp(ds->value->ptr, con->uri.path->ptr, ulen)
+            || (ds->value->ptr[ulen] != '\0'
+                && ds->value->ptr[ulen] != '/'
+                && ds->value->ptr[ulen] != '?'))
+        && 0 == blen
+        && !(con->parsed_response & HTTP_STATUS) /*no "Status" or NPH response*/
+        && 1 == con->response.headers->used) {
+        if (++con->loops_per_request > 5) {
+            log_error_write(srv, __FILE__, __LINE__, "sb",
+                            "too many internal loops while processing request:",
+                            con->request.orig_uri);
+            con->http_status = 500; /* Internal Server Error */
+            con->mode = DIRECT;
+            return HANDLER_FINISHED;
+        }
+
+        buffer_copy_buffer(con->request.uri, ds->value);
+
+        if (con->request.content_length) {
+            if (con->request.content_length
+                != con->request_content_queue->bytes_in) {
+                con->keep_alive = 0;
+            }
+            con->request.content_length = 0;
+            chunkqueue_reset(con->request_content_queue);
+        }
+
+        if (con->http_status != 307 && con->http_status != 308) {
+            /* Note: request body (if any) sent to initial dynamic handler
+             * and is not available to the internal redirect */
+            con->request.http_method = HTTP_METHOD_GET;
+        }
+
+        /*(caller must reset request as follows)*/
+        /*connection_response_reset(srv, con);*/ /*(sets con->http_status = 0)*/
+        /*plugins_call_connection_reset(srv, con);*/
+
+        return HANDLER_COMEBACK;
+    }
+
+    return HANDLER_GO_ON;
+}
+
+
+static int http_response_process_headers(server *srv, connection *con, http_response_opts *opts, buffer *hdrs) {
+    char *ns;
+    const char *s;
+    int line = 0;
+
+    for (s = hdrs->ptr; NULL != (ns = strchr(s, '\n')); s = ns + 1, ++line) {
+        const char *key, *value;
+        int key_len;
+        data_string *ds;
+
+        /* strip the \n */
+        ns[0] = '\0';
+        if (ns > s && ns[-1] == '\r') ns[-1] = '\0';
+
+        if (0 == line && 0 == strncmp(s, "HTTP/1.", 7)) {
+            /* non-parsed headers ... we parse them anyway */
+            if ((s[7] == '1' || s[7] == '0') && s[8] == ' ') {
+                /* after the space should be a status code for us */
+                int status = strtol(s+9, NULL, 10);
+                if (status >= 100 && status < 1000) {
+                    con->parsed_response |= HTTP_STATUS;
+                    con->http_status = status;
+                } /* else we expected 3 digits and didn't get them */
+            }
+
+            if (0 == con->http_status) {
+                log_error_write(srv, __FILE__, __LINE__, "ss",
+                                "invalid HTTP status line:", s);
+                con->http_status = 502; /* Bad Gateway */
+                con->mode = DIRECT;
+                return -1;
+            }
+
+            continue;
+        }
+
+        /* parse the headers */
+        key = s;
+        if (NULL == (value = strchr(s, ':'))) {
+            /* we expect: "<key>: <value>\r\n" */
+            continue;
+        }
+
+        key_len = value - key;
+        do { ++value; } while (*value == ' ' || *value == '\t'); /* skip LWS */
+
+        if (opts->authorizer) {
+            if (0 == con->http_status || 200 == con->http_status) {
+                if (key_len == 6 && 0 == strncasecmp(key, "Status", key_len)) {
+                    int status = strtol(value, NULL, 10);
+                    if (status >= 100 && status < 1000) {
+                        con->http_status = status;
+                    } else {
+                        con->http_status = 502; /* Bad Gateway */
+                        break;
+                    }
+                } else if (key_len > 9
+                           && 0==strncasecmp(key, CONST_STR_LEN("Variable-"))) {
+                    ds = (data_string *)
+                      array_get_unused_element(con->environment, TYPE_STRING);
+                    if (NULL == ds) ds = data_string_init();
+                    buffer_copy_string_len(ds->key, key + 9, key_len - 9);
+                    buffer_copy_string(ds->value, value);
+
+                    array_insert_unique(con->environment, (data_unset *)ds);
+                }
+                continue;
+            }
+        }
+
+        switch(key_len) {
+        case 4:
+            if (0 == strncasecmp(key, "Date", key_len)) {
+                con->parsed_response |= HTTP_DATE;
+            }
+            break;
+        case 6:
+            if (0 == strncasecmp(key, "Status", key_len)) {
+                int status;
+                if (opts->backend == BACKEND_PROXY) break; /*(pass w/o parse)*/
+                status = strtol(value, NULL, 10);
+                if (status >= 100 && status < 1000) {
+                    con->http_status = status;
+                    con->parsed_response |= HTTP_STATUS;
+                } else {
+                    con->http_status = 502;
+                    con->mode = DIRECT;
+                }
+                continue; /* do not send Status to client */
+            }
+            break;
+        case 7:
+            if (0 == strncasecmp(key, "Upgrade", key_len)) {
+                /*(technically, should also verify Connection: upgrade)*/
+                /*(flag only for mod_proxy and mod_cgi (for now))*/
+                if (opts->backend == BACKEND_PROXY
+                    || opts->backend == BACKEND_CGI) {
+                    con->parsed_response |= HTTP_UPGRADE;
+                }
+            }
+            break;
+        case 8:
+            if (0 == strncasecmp(key, "Location", key_len)) {
+                con->parsed_response |= HTTP_LOCATION;
+            }
+            break;
+        case 10:
+            if (0 == strncasecmp(key, "Connection", key_len)) {
+                if (opts->backend == BACKEND_PROXY) continue;
+                con->response.keep_alive =
+                  (0 == strcasecmp(value, "Keep-Alive")) ? 1 : 0;
+                con->parsed_response |= HTTP_CONNECTION;
+            }
+            else if (0 == strncasecmp(key, "Set-Cookie", key_len)) {
+                con->parsed_response |= HTTP_SET_COOKIE;
+            }
+            break;
+        case 14:
+            if (0 == strncasecmp(key, "Content-Length", key_len)) {
+                con->response.content_length = strtoul(value, NULL, 10);
+                con->parsed_response |= HTTP_CONTENT_LENGTH;
+            }
+            break;
+        case 16:
+            if (0 == strncasecmp(key, "Content-Location", key_len)) {
+                con->parsed_response |= HTTP_CONTENT_LOCATION;
+            }
+            break;
+        case 17:
+            if (0 == strncasecmp(key, "Transfer-Encoding", key_len)) {
+                if (opts->backend == BACKEND_PROXY) continue;
+                con->parsed_response |= HTTP_TRANSFER_ENCODING;
+            }
+            break;
+        default:
+            break;
+        }
+
+        ds = (data_string *)
+          array_get_unused_element(con->response.headers, TYPE_STRING);
+        if (NULL == ds) ds = data_response_init();
+        buffer_copy_string_len(ds->key, key, key_len);
+        buffer_copy_string(ds->value, value);
+
+        array_insert_unique(con->response.headers, (data_unset *)ds);
+    }
+
+    /* CGI/1.1 rev 03 - 7.2.1.2 */
+    /* (proxy requires Status-Line, so never true for proxy)*/
+    if ((con->parsed_response & HTTP_LOCATION) &&
+        !(con->parsed_response & HTTP_STATUS)) {
+        con->http_status = 302;
+    }
+
+    return 0;
+}
+
+
+handler_t http_response_parse_headers(server *srv, connection *con, http_response_opts *opts, buffer *b) {
+    /**
+     * possible formats of response headers:
+     *
+     * proxy or NPH (non-parsed headers):
+     *
+     *   HTTP/1.0 200 Ok\n
+     *   Header: Value\n
+     *   \n
+     *
+     * CGI:
+     *
+     *   Header: Value\n
+     *   Status: 200\n
+     *   \n
+     *
+     * and different mixes of \n and \r\n combinations
+     *
+     * Some users also forget about CGI and just send a response
+     * and hope we handle it. No headers, no header-content separator
+     */
+
+    int is_nph = (0 == strncmp(b->ptr, "HTTP/1.", 7)); /*nph (non-parsed hdrs)*/
+    int is_header_end = 0;
+    size_t last_eol = 0;
+    size_t i = 0, header_len = buffer_string_length(b);
+    const char *bstart;
+    size_t blen;
+
+    if (b->ptr[0] == '\n' || (b->ptr[0] == '\r' && b->ptr[1] == '\n')) {
+        /* no HTTP headers */
+        i = (b->ptr[0] == '\n') ? 1 : 2;
+        is_header_end = 1;
+    } else if (is_nph || b->ptr[(i = strcspn(b->ptr, ":\n"))] == ':') {
+        /* HTTP headers */
+        ++i;
+        for (char *c; NULL != (c = strchr(b->ptr+i, '\n')); ++i) {
+            i = (uintptr_t)(c - b->ptr);
+            /**
+             * check if we saw a \n(\r)?\n sequence
+             */
+            if (last_eol > 0 &&
+                ((i - last_eol == 1) ||
+                 (i - last_eol == 2 && b->ptr[i - 1] == '\r'))) {
+                is_header_end = 1;
+                break;
+            }
+
+            last_eol = i;
+        }
+    } else if (opts->backend == BACKEND_CGI) {
+        /* no HTTP headers, but a body (special-case for CGI compat) */
+        /* no colon found; does not appear to be HTTP headers */
+        if (0 != http_chunk_append_buffer(srv, con, b)) {
+            return HANDLER_ERROR;
+        }
+        con->http_status = 200; /* OK */
+        con->file_started = 1;
+        return HANDLER_GO_ON;
+    } else {
+        /* invalid response headers */
+        con->http_status = 502; /* Bad Gateway */
+        con->mode = DIRECT;
+        return HANDLER_FINISHED;
+    }
+
+    if (!is_header_end) {
+        /*(reuse MAX_HTTP_REQUEST_HEADER as max size
+         * for response headers from backends)*/
+        if (header_len > MAX_HTTP_REQUEST_HEADER) {
+            log_error_write(srv, __FILE__, __LINE__, "sb",
+                            "response headers too large for", con->uri.path);
+            con->http_status = 502; /* Bad Gateway */
+            con->mode = DIRECT;
+            return HANDLER_FINISHED;
+        }
+        return HANDLER_GO_ON;
+    }
+
+    /* the body starts after the EOL */
+    bstart = b->ptr + (i + 1);
+    blen = header_len - (i + 1);
+
+    /* strip the last \r?\n */
+    if (i > 0 && (b->ptr[i - 1] == '\r')) {
+        i--;
+    }
+
+    buffer_string_set_length(b, i);
+
+    if (opts->backend == BACKEND_PROXY && !is_nph) {
+        /* invalid response Status-Line from HTTP proxy */
+        con->http_status = 502; /* Bad Gateway */
+        con->mode = DIRECT;
+        return HANDLER_FINISHED;
+    }
+
+    if (0 != http_response_process_headers(srv, con, opts, b)) {
+        return HANDLER_ERROR;
+    }
+
+    con->file_started = 1;
+
+    if (opts->authorizer
+        && (con->http_status == 0 || con->http_status == 200)) {
+        return HANDLER_GO_ON;
+    }
+
+    if (con->mode == DIRECT) {
+        return HANDLER_FINISHED;
+    }
+
+    if (opts->local_redir && con->http_status >= 300 && con->http_status < 400){
+        /*(con->parsed_response & HTTP_LOCATION)*/
+        handler_t rc = http_response_process_local_redir(srv, con, blen);
+        if (con->mode == DIRECT) con->file_started = 0;
+        if (rc != HANDLER_GO_ON) return rc;
+    }
+
+    if (opts->xsendfile_allow) {
+        data_string *ds;
+        /* X-Sendfile2 is deprecated; historical for fastcgi */
+        if (opts->backend == BACKEND_FASTCGI
+            && NULL != (ds = (data_string *) array_get_element(con->response.headers, "X-Sendfile2"))) {
+            http_response_xsendfile2(srv, con, ds->value, opts->xsendfile_docroot);
+            buffer_reset(ds->value); /*(do not send to client)*/
+            if (con->mode == DIRECT) con->file_started = 0;
+            return HANDLER_FINISHED;
+        } else if (NULL != (ds = (data_string *) array_get_element(con->response.headers, "X-Sendfile"))
+                   || (opts->backend == BACKEND_FASTCGI /* X-LIGHTTPD-send-file is deprecated; historical for fastcgi */
+                       && NULL != (ds = (data_string *) array_get_element(con->response.headers, "X-LIGHTTPD-send-file")))) {
+            http_response_xsendfile(srv, con, ds->value, opts->xsendfile_docroot);
+            buffer_reset(ds->value); /*(do not send to client)*/
+            if (con->mode == DIRECT) con->file_started = 0;
+            return HANDLER_FINISHED;
+        }
+    }
+
+    if (blen > 0) {
+        if (0 != http_chunk_append_mem(srv, con, bstart, blen)) {
+            return HANDLER_ERROR;
+        }
+    }
+
+    /* (callback for response headers complete) */
+    return (opts->headers) ? opts->headers(srv, con, opts) : HANDLER_GO_ON;
+}
+
+
+handler_t http_response_read(server *srv, connection *con, http_response_opts *opts, buffer *b, int fd, int *fde_ndx) {
+    while (1) {
+        ssize_t n;
+        size_t avail = buffer_string_space(b);
+        unsigned int toread = 4096;
+
+        if (0 == fdevent_ioctl_fionread(fd, opts->fdfmt, (int *)&toread)) {
+            if (avail < toread) {
+                if (toread < 4096)
+                    toread = 4096;
+                else if (toread > MAX_READ_LIMIT)
+                    toread = MAX_READ_LIMIT;
+            }
+            else if (0 == toread) {
+              #if 0
+                return (fdevent_event_get_interest(srv->ev, fd) & FDEVENT_IN)
+                  ? HANDLER_FINISHED  /* read finished */
+                  : HANDLER_GO_ON;    /* optimistic read; data not ready */
+              #else
+                if (!(fdevent_event_get_interest(srv->ev, fd) & FDEVENT_IN))
+                    return HANDLER_GO_ON; /* optimistic read; data not ready */
+                toread = 4096; /* let read() below indicate if EOF or EAGAIN */
+              #endif
+            }
+        }
+
+        if (con->conf.stream_response_body & FDEVENT_STREAM_RESPONSE_BUFMIN) {
+            off_t cqlen = chunkqueue_length(con->write_queue);
+            if (cqlen + (off_t)toread > 65536 - 4096) {
+                if (!con->is_writable) {
+                    /*(defer removal of FDEVENT_IN interest since
+                     * connection_state_machine() might be able to send data
+                     * immediately, unless !con->is_writable, where
+                     * connection_state_machine() might not loop back to call
+                     * mod_proxy_handle_subrequest())*/
+                    fdevent_event_clr(srv->ev, fde_ndx, fd, FDEVENT_IN);
+                }
+                if (cqlen >= 65536-1) return HANDLER_GO_ON;
+                toread = 65536 - 1 - (unsigned int)cqlen;
+                /* Note: heuristic is fuzzy in that it limits how much to read
+                 * from backend based on how much is pending to write to client.
+                 * Modules where data from backend is framed (e.g. FastCGI) may
+                 * want to limit how much is buffered from backend while waiting
+                 * for a complete data frame or data packet from backend. */
+            }
+        }
+
+        if (avail < toread) {
+            /*(add avail+toread to reduce allocations when ioctl EOPNOTSUPP)*/
+            avail = avail ? avail - 1 + toread : toread;
+            buffer_string_prepare_append(b, avail);
+        }
+
+        n = read(fd, b->ptr+buffer_string_length(b), avail);
+
+        if (n < 0) {
+            switch (errno) {
+              case EAGAIN:
+             #ifdef EWOULDBLOCK
+             #if EWOULDBLOCK != EAGAIN
+              case EWOULDBLOCK:
+             #endif
+             #endif
+              case EINTR:
+                return HANDLER_GO_ON;
+              default:
+                log_error_write(srv, __FILE__, __LINE__, "ssdd",
+                                "read():", strerror(errno), con->fd, fd);
+                return HANDLER_ERROR;
+            }
+        }
+
+        buffer_commit(b, (size_t)n);
+
+        if (NULL != opts->parse) {
+            handler_t rc = opts->parse(srv, con, opts, b, (size_t)n);
+            if (rc != HANDLER_GO_ON) return rc;
+        } else if (0 == n) {
+            /* note: no further data is sent to backend after read EOF on socket
+             * (not checking for half-closed TCP socket)
+             * (backend should read all data desired prior to closing socket,
+             *  though might send app-level close data frame, if applicable) */
+            return HANDLER_FINISHED; /* read finished */
+        } else if (0 == con->file_started) {
+            /* split header from body */
+            handler_t rc = http_response_parse_headers(srv, con, opts, b);
+            if (rc != HANDLER_GO_ON) return rc;
+            /* accumulate response in b until headers completed (or error) */
+            if (con->file_started) buffer_string_set_length(b, 0);
+        } else {
+            if (0 != http_chunk_append_buffer(srv, con, b)) {
+                /* error writing to tempfile;
+                 * truncate response or send 500 if nothing sent yet */
+                return HANDLER_ERROR;
+            }
+            buffer_string_set_length(b, 0);
+        }
+
+        if ((con->conf.stream_response_body & FDEVENT_STREAM_RESPONSE_BUFMIN)
+            && chunkqueue_length(con->write_queue) > 65536 - 4096) {
+            if (!con->is_writable) {
+                /*(defer removal of FDEVENT_IN interest since
+                 * connection_state_machine() might be able to send
+                 * data immediately, unless !con->is_writable, where
+                 * connection_state_machine() might not loop back to
+                 * call the subrequest handler)*/
+                fdevent_event_clr(srv->ev, fde_ndx, fd, FDEVENT_IN);
+            }
+            break;
+        }
+
+        if ((size_t)n < avail)
+            break; /* emptied kernel read buffer or partial read */
+    }
+
+    return HANDLER_GO_ON;
+}
+
+
 int http_cgi_headers (server *srv, connection *con, http_cgi_opts *opts, http_cgi_header_append_cb cb, void *vdata) {
 
     /* CGI-SPEC 6.1.2, FastCGI spec 6.3 and SCGI spec */
 
     int rc = 0;
-    unsigned short port;
     server_socket *srv_sock = con->srv_socket;
     const char *s;
     size_t n;
     char buf[LI_ITOSTRING_LENGTH];
-  #ifdef HAVE_IPV6
-    char b2[INET6_ADDRSTRLEN + 1];
-  #endif
     sock_addr *addr;
     sock_addr addrbuf;
+    char b2[INET6_ADDRSTRLEN + 1];
 
     /* (CONTENT_LENGTH must be first for SCGI) */
     if (!opts->authorizer) {
@@ -902,40 +1449,22 @@ int http_cgi_headers (server *srv, connection *con, http_cgi_opts *opts, http_cg
     rc |= cb(vdata, CONST_STR_LEN("GATEWAY_INTERFACE"),
                     CONST_STR_LEN("CGI/1.1"));
 
+    rc |= cb(vdata, CONST_STR_LEN("REQUEST_SCHEME"),
+                    CONST_BUF_LEN(con->uri.scheme));
+
     if (buffer_is_equal_caseless_string(con->uri.scheme,
                                         CONST_STR_LEN("https"))) {
         rc |= cb(vdata, CONST_STR_LEN("HTTPS"), CONST_STR_LEN("on"));
     }
 
     addr = &srv_sock->addr;
-  #ifdef HAVE_IPV6
-    port = addr->plain.sa_family == AF_INET6
-         ? addr->ipv6.sin6_port
-         : addr->ipv4.sin_port;
-  #else
-    port = addr->ipv4.sin_port;
-  #endif
-    li_utostrn(buf, sizeof(buf), ntohs(port));
+    li_utostrn(buf, sizeof(buf), sock_addr_get_port(addr));
     rc |= cb(vdata, CONST_STR_LEN("SERVER_PORT"), buf, strlen(buf));
 
     switch (addr->plain.sa_family) {
-  #ifdef HAVE_IPV6
-    case AF_INET6:
-        if (0 ==memcmp(&addr->ipv6.sin6_addr,&in6addr_any,sizeof(in6addr_any))){
-            socklen_t addrlen = sizeof(addrbuf);
-            if (0 == getsockname(con->fd,(struct sockaddr *)&addrbuf,&addrlen)){
-                addr = &addrbuf;
-            } else {
-                s = "";
-                break;
-            }
-        }
-        s = inet_ntop(AF_INET6, (const void *) &(addr->ipv6.sin6_addr),
-                      b2, sizeof(b2)-1);
-        break;
-  #endif
     case AF_INET:
-        if (srv_sock->addr.ipv4.sin_addr.s_addr == INADDR_ANY) {
+    case AF_INET6:
+        if (sock_addr_is_addr_wildcard(addr)) {
             socklen_t addrlen = sizeof(addrbuf);
             if (0 == getsockname(con->fd,(struct sockaddr *)&addrbuf,&addrlen)){
                 addr = &addrbuf;
@@ -944,12 +1473,8 @@ int http_cgi_headers (server *srv, connection *con, http_cgi_opts *opts, http_cg
                 break;
             }
         }
-      #ifdef HAVE_IPV6
-        s = inet_ntop(AF_INET, (const void *) &(addr->ipv4.sin_addr),
-                      b2, sizeof(b2)-1);
-      #else
-        s = inet_ntoa(addr->ipv4.sin_addr);
-      #endif
+        s = sock_addr_inet_ntop(addr, b2, sizeof(b2)-1);
+        if (NULL == s) s = "";
         break;
     default:
         s = "";
@@ -979,19 +1504,12 @@ int http_cgi_headers (server *srv, connection *con, http_cgi_opts *opts, http_cg
     rc |= cb(vdata, CONST_STR_LEN("REMOTE_ADDR"),
                     CONST_BUF_LEN(con->dst_addr_buf));
 
-  #ifdef HAVE_IPV6
-    port = con->dst_addr.plain.sa_family == AF_INET6
-         ? con->dst_addr.ipv6.sin6_port
-         : con->dst_addr.ipv4.sin_port;
-  #else
-    port = con->dst_addr.ipv4.sin_port;
-  #endif
-    li_utostrn(buf, sizeof(buf), ntohs(port));
+    li_utostrn(buf, sizeof(buf), sock_addr_get_port(&con->dst_addr));
     rc |= cb(vdata, CONST_STR_LEN("REMOTE_PORT"), buf, strlen(buf));
 
     for (n = 0; n < con->request.headers->used; n++) {
         data_string *ds = (data_string *)con->request.headers->data[n];
-        if (!buffer_is_empty(ds->value) && !buffer_is_empty(ds->key)) {
+        if (!buffer_string_is_empty(ds->value) && !buffer_is_empty(ds->key)) {
             /* Security: Do not emit HTTP_PROXY in environment.
              * Some executables use HTTP_PROXY to configure
              * outgoing proxy.  See also https://httpoxy.org/ */
@@ -1006,9 +1524,7 @@ int http_cgi_headers (server *srv, connection *con, http_cgi_opts *opts, http_cg
         }
     }
 
-  #ifdef USE_OPENSSL
-    if (con->ssl) http_cgi_ssl_env(srv, con);
-  #endif
+    srv->request_env(srv, con);
 
     for (n = 0; n < con->environment->used; n++) {
         data_string *ds = (data_string *)con->environment->data[n];
@@ -1022,37 +1538,3 @@ int http_cgi_headers (server *srv, connection *con, http_cgi_opts *opts, http_cg
 
     return rc;
 }
-
-
-#ifdef USE_OPENSSL
-void http_cgi_ssl_env(server *srv, connection *con) {
-    const char *s;
-    const SSL_CIPHER *cipher;
-    UNUSED(srv);
-
-    if (!con->ssl) return;
-
-    s = SSL_get_version(con->ssl);
-    array_set_key_value(con->environment,
-                        CONST_STR_LEN("SSL_PROTOCOL"),
-                        s, strlen(s));
-
-    if ((cipher = SSL_get_current_cipher(con->ssl))) {
-        int usekeysize, algkeysize;
-        char buf[LI_ITOSTRING_LENGTH];
-        s = SSL_CIPHER_get_name(cipher);
-        array_set_key_value(con->environment,
-                            CONST_STR_LEN("SSL_CIPHER"),
-                            s, strlen(s));
-        usekeysize = SSL_CIPHER_get_bits(cipher, &algkeysize);
-        li_itostrn(buf, sizeof(buf), usekeysize);
-        array_set_key_value(con->environment,
-                            CONST_STR_LEN("SSL_CIPHER_USEKEYSIZE"),
-                            buf, strlen(buf));
-        li_itostrn(buf, sizeof(buf), algkeysize);
-        array_set_key_value(con->environment,
-                            CONST_STR_LEN("SSL_CIPHER_ALGKEYSIZE"),
-                            buf, strlen(buf));
-    }
-}
-#endif
