@@ -48,6 +48,14 @@ void http_auth_backend_set (const http_auth_backend_t *backend)
     memcpy(http_auth_backends+i, backend, sizeof(http_auth_backend_t));
 }
 
+
+void http_auth_dumbdata_reset (void)
+{
+    memset(http_auth_schemes, 0, sizeof(http_auth_schemes));
+    memset(http_auth_backends, 0, sizeof(http_auth_backends));
+}
+
+
 http_auth_require_t * http_auth_require_init (void)
 {
     http_auth_require_t *require = calloc(1, sizeof(http_auth_require_t));
