@@ -6,10 +6,8 @@
 
 #include "plugin.h"
 
-#include <ctype.h>
 #include <stdlib.h>
 #include <string.h>
-#include <stdio.h>
 
 /* plugin config for all request/connections */
 typedef struct {
@@ -89,6 +87,13 @@ SETDEFAULTS_FUNC(mod_alias_set_defaults) {
 		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
 			return HANDLER_ERROR;
 		}
+
+		if (!array_is_kvstring(s->alias)) {
+			log_error_write(srv, __FILE__, __LINE__, "s",
+					"unexpected value for alias.url; expected list of \"urlpath\" => \"filepath\"");
+			return HANDLER_ERROR;
+		}
+
 		if (s->alias->used >= 2) {
 			const array *a = s->alias;
 			size_t j, k;
