@@ -13,27 +13,17 @@
 
 #include <sys/types.h>
 #include "sys-mmap.h"
-
-#ifdef __WIN32
-# include <winsock2.h>
-#else
-# include <sys/socket.h>
+#include "sys-socket.h"
 # include <sys/wait.h>
-# include <netinet/in.h>
-# include <arpa/inet.h>
-#endif
 
 #include <unistd.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
 #include <fdevent.h>
-#include <signal.h>
-#include <ctype.h>
-#include <assert.h>
 
-#include <stdio.h>
 #include <fcntl.h>
+#include <signal.h>
 
 static int pipe_cloexec(int pipefd[2]) {
   #ifdef HAVE_PIPE2
@@ -48,8 +38,6 @@ static int pipe_cloexec(int pipefd[2]) {
       : -1;
 }
 
-enum {EOL_UNSET, EOL_N, EOL_RN};
-
 typedef struct {
 	char **ptr;
 
@@ -58,7 +46,7 @@ typedef struct {
 } char_array;
 
 typedef struct {
-	pid_t *ptr;
+	struct { pid_t pid; void *ctx; } *ptr;
 	size_t used;
 	size_t size;
 } buffer_pid_t;
@@ -66,7 +54,9 @@ typedef struct {
 typedef struct {
 	array *cgi;
 	unsigned short execute_x_only;
+	unsigned short local_redir;
 	unsigned short xsendfile_allow;
+	unsigned short upgrade;
 	array *xsendfile_docroot;
 } plugin_config;
 
@@ -74,8 +64,6 @@ typedef struct {
 	PLUGIN_DATA;
 	buffer_pid_t cgi_pid;
 
-	buffer *parse_response;
-
 	plugin_config **config_storage;
 
 	plugin_config conf;
@@ -92,8 +80,8 @@ typedef struct {
 	plugin_data *plugin_data; /* dumb pointer */
 
 	buffer *response;
-	buffer *response_header;
 	buffer *cgi_handler;      /* dumb pointer */
+	http_response_opts opts;
 	plugin_config conf;
 } handler_ctx;
 
@@ -103,7 +91,6 @@ static handler_ctx * cgi_handler_ctx_init(void) {
 	force_assert(hctx);
 
 	hctx->response = buffer_init();
-	hctx->response_header = buffer_init();
 	hctx->fd = -1;
 	hctx->fdtocgi = -1;
 
@@ -112,13 +99,9 @@ static handler_ctx * cgi_handler_ctx_init(void) {
 
 static void cgi_handler_ctx_free(handler_ctx *hctx) {
 	buffer_free(hctx->response);
-	buffer_free(hctx->response_header);
-
 	free(hctx);
 }
 
-enum {FDEVENT_HANDLED_UNSET, FDEVENT_HANDLED_FINISHED, FDEVENT_HANDLED_NOT_FINISHED, FDEVENT_HANDLED_COMEBACK, FDEVENT_HANDLED_ERROR};
-
 INIT_FUNC(mod_cgi_init) {
 	plugin_data *p;
 
@@ -126,8 +109,6 @@ INIT_FUNC(mod_cgi_init) {
 
 	force_assert(p);
 
-	p->parse_response = buffer_init();
-
 	return p;
 }
 
@@ -156,8 +137,6 @@ FREE_FUNC(mod_cgi_free) {
 
 	if (r->ptr) free(r->ptr);
 
-	buffer_free(p->parse_response);
-
 	free(p);
 
 	return HANDLER_GO_ON;
@@ -172,6 +151,8 @@ SETDEFAULTS_FUNC(mod_fastcgi_set_defaults) {
 		{ "cgi.execute-x-only",          NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION },     /* 1 */
 		{ "cgi.x-sendfile",              NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION },     /* 2 */
 		{ "cgi.x-sendfile-docroot",      NULL, T_CONFIG_ARRAY,   T_CONFIG_SCOPE_CONNECTION },     /* 3 */
+		{ "cgi.local-redir",             NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION },     /* 4 */
+		{ "cgi.upgrade",                 NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION },     /* 5 */
 		{ NULL,                          NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET}
 	};
 
@@ -189,13 +170,17 @@ SETDEFAULTS_FUNC(mod_fastcgi_set_defaults) {
 
 		s->cgi    = array_init();
 		s->execute_x_only = 0;
+		s->local_redir    = 0;
 		s->xsendfile_allow= 0;
 		s->xsendfile_docroot = array_init();
+		s->upgrade        = 0;
 
 		cv[0].destination = s->cgi;
 		cv[1].destination = &(s->execute_x_only);
 		cv[2].destination = &(s->xsendfile_allow);
 		cv[3].destination = s->xsendfile_docroot;
+		cv[4].destination = &(s->local_redir);
+		cv[5].destination = &(s->upgrade);
 
 		p->config_storage[i] = s;
 
@@ -203,6 +188,12 @@ SETDEFAULTS_FUNC(mod_fastcgi_set_defaults) {
 			return HANDLER_ERROR;
 		}
 
+		if (!array_is_kvstring(s->cgi)) {
+			log_error_write(srv, __FILE__, __LINE__, "s",
+					"unexpected value for cgi.assign; expected list of \"ext\" => \"exepath\"");
+			return HANDLER_ERROR;
+		}
+
 		if (s->xsendfile_docroot->used) {
 			size_t j;
 			for (j = 0; j < s->xsendfile_docroot->used; ++j) {
@@ -227,17 +218,9 @@ SETDEFAULTS_FUNC(mod_fastcgi_set_defaults) {
 }
 
 
-static int cgi_pid_add(server *srv, plugin_data *p, pid_t pid) {
-	int m = -1;
-	size_t i;
+static void cgi_pid_add(plugin_data *p, pid_t pid, void *ctx) {
 	buffer_pid_t *r = &(p->cgi_pid);
 
-	UNUSED(srv);
-
-	for (i = 0; i < r->used; i++) {
-		if (r->ptr[i] > m) m = r->ptr[i];
-	}
-
 	if (r->size == 0) {
 		r->size = 16;
 		r->ptr = malloc(sizeof(*r->ptr) * r->size);
@@ -248,391 +231,44 @@ static int cgi_pid_add(server *srv, plugin_data *p, pid_t pid) {
 		force_assert(r->ptr);
 	}
 
-	r->ptr[r->used++] = pid;
-
-	return m;
+	r->ptr[r->used].pid = pid;
+	r->ptr[r->used].ctx = ctx;
+	++r->used;
 }
 
-static int cgi_pid_del(server *srv, plugin_data *p, pid_t pid) {
-	size_t i;
+static void cgi_pid_kill(plugin_data *p, pid_t pid) {
+    buffer_pid_t *r = &(p->cgi_pid);
+    for (size_t i = 0; i < r->used; ++i) {
+        if (r->ptr[i].pid == pid) {
+            r->ptr[i].ctx = NULL;
+            kill(pid, SIGTERM);
+            return;
+        }
+    }
+}
+
+static void cgi_pid_del(plugin_data *p, size_t i) {
 	buffer_pid_t *r = &(p->cgi_pid);
 
-	UNUSED(srv);
-
-	for (i = 0; i < r->used; i++) {
-		if (r->ptr[i] == pid) break;
-	}
-
-	if (i != r->used) {
-		/* found */
-
 		if (i != r->used - 1) {
 			r->ptr[i] = r->ptr[r->used - 1];
 		}
 		r->used--;
-	}
-
-	return 0;
 }
 
-static int cgi_response_parse(server *srv, connection *con, plugin_data *p, buffer *in) {
-	char *ns;
-	const char *s;
-	int line = 0;
-
-	UNUSED(srv);
-
-	buffer_copy_buffer(p->parse_response, in);
-
-	for (s = p->parse_response->ptr;
-	     NULL != (ns = strchr(s, '\n'));
-	     s = ns + 1, line++) {
-		const char *key, *value;
-		int key_len;
-		data_string *ds;
-
-		/* strip the \n */
-		ns[0] = '\0';
-
-		if (ns > s && ns[-1] == '\r') ns[-1] = '\0';
-
-		if (line == 0 &&
-		    0 == strncmp(s, "HTTP/1.", 7)) {
-			/* non-parsed header ... we parse them anyway */
-
-			if ((s[7] == '1' ||
-			     s[7] == '0') &&
-			    s[8] == ' ') {
-				int status;
-				/* after the space should be a status code for us */
-
-				status = strtol(s+9, NULL, 10);
-
-				if (status >= 100 &&
-				    status < 1000) {
-					/* we expected 3 digits and didn't got them */
-					con->parsed_response |= HTTP_STATUS;
-					con->http_status = status;
-				}
-			}
-		} else {
-			/* parse the headers */
-			key = s;
-			if (NULL == (value = strchr(s, ':'))) {
-				/* we expect: "<key>: <value>\r\n" */
-				continue;
-			}
-
-			key_len = value - key;
-			value += 1;
-
-			/* skip LWS */
-			while (*value == ' ' || *value == '\t') value++;
-
-			if (NULL == (ds = (data_string *)array_get_unused_element(con->response.headers, TYPE_STRING))) {
-				ds = data_response_init();
-			}
-			buffer_copy_string_len(ds->key, key, key_len);
-			buffer_copy_string(ds->value, value);
-
-			array_insert_unique(con->response.headers, (data_unset *)ds);
-
-			switch(key_len) {
-			case 4:
-				if (0 == strncasecmp(key, "Date", key_len)) {
-					con->parsed_response |= HTTP_DATE;
-				}
-				break;
-			case 6:
-				if (0 == strncasecmp(key, "Status", key_len)) {
-					int status = strtol(value, NULL, 10);
-					if (status >= 100 && status < 1000) {
-						con->http_status = status;
-						con->parsed_response |= HTTP_STATUS;
-					} else {
-						con->http_status = 502;
-					}
-				}
-				break;
-			case 8:
-				if (0 == strncasecmp(key, "Location", key_len)) {
-					con->parsed_response |= HTTP_LOCATION;
-				}
-				break;
-			case 10:
-				if (0 == strncasecmp(key, "Connection", key_len)) {
-					con->response.keep_alive = (0 == strcasecmp(value, "Keep-Alive")) ? 1 : 0;
-					con->parsed_response |= HTTP_CONNECTION;
-				}
-				break;
-			case 14:
-				if (0 == strncasecmp(key, "Content-Length", key_len)) {
-					con->response.content_length = strtoul(value, NULL, 10);
-					con->parsed_response |= HTTP_CONTENT_LENGTH;
-				}
-				break;
-			default:
-				break;
-			}
-		}
-	}
-
-	/* CGI/1.1 rev 03 - 7.2.1.2 */
-	if ((con->parsed_response & HTTP_LOCATION) &&
-	    !(con->parsed_response & HTTP_STATUS)) {
-		con->http_status = 302;
-	}
-
-	return 0;
-}
-
-
-static int cgi_demux_response(server *srv, handler_ctx *hctx) {
-	plugin_data *p    = hctx->plugin_data;
-	connection  *con  = hctx->remote_conn;
-
-	while(1) {
-		int n;
-		int toread;
-
-#if defined(__WIN32)
-		buffer_string_prepare_copy(hctx->response, 4 * 1024);
-#else
-		if (ioctl(hctx->fd, FIONREAD, &toread) || toread <= 4*1024) {
-			buffer_string_prepare_copy(hctx->response, 4 * 1024);
-		} else {
-			if (toread > MAX_READ_LIMIT) toread = MAX_READ_LIMIT;
-			buffer_string_prepare_copy(hctx->response, toread);
-		}
-#endif
-
-		if (-1 == (n = read(hctx->fd, hctx->response->ptr, hctx->response->size - 1))) {
-			if (errno == EAGAIN || errno == EINTR) {
-				/* would block, wait for signal */
-				fdevent_event_add(srv->ev, &(hctx->fde_ndx), hctx->fd, FDEVENT_IN);
-				return FDEVENT_HANDLED_NOT_FINISHED;
-			}
-			/* error */
-			log_error_write(srv, __FILE__, __LINE__, "sdd", strerror(errno), con->fd, hctx->fd);
-			return FDEVENT_HANDLED_ERROR;
-		}
-
-		if (n == 0) {
-			/* read finished */
-			return FDEVENT_HANDLED_FINISHED;
-		}
-
-		buffer_commit(hctx->response, n);
-
-		/* split header from body */
-
-		if (con->file_started == 0) {
-			int is_header = 0;
-			int is_header_end = 0;
-			size_t last_eol = 0;
-			size_t i, header_len;
-
-			buffer_append_string_buffer(hctx->response_header, hctx->response);
-
-			/**
-			 * we have to handle a few cases:
-			 *
-			 * nph:
-			 * 
-			 *   HTTP/1.0 200 Ok\n
-			 *   Header: Value\n
-			 *   \n
-			 *
-			 * CGI:
-			 *   Header: Value\n
-			 *   Status: 200\n
-			 *   \n
-			 *
-			 * and different mixes of \n and \r\n combinations
-			 * 
-			 * Some users also forget about CGI and just send a response and hope 
-			 * we handle it. No headers, no header-content seperator
-			 * 
-			 */
-			
-			/* nph (non-parsed headers) */
-			if (0 == strncmp(hctx->response_header->ptr, "HTTP/1.", 7)) is_header = 1;
-
-			header_len = buffer_string_length(hctx->response_header);
-			for (i = 0; !is_header_end && i < header_len; i++) {
-				char c = hctx->response_header->ptr[i];
-
-				switch (c) {
-				case ':':
-					/* we found a colon
-					 *
-					 * looks like we have a normal header 
-					 */
-					is_header = 1;
-					break;
-				case '\n':
-					/* EOL */
-					if (is_header == 0) {
-						/* we got a EOL but we don't seem to got a HTTP header */
-
-						is_header_end = 1;
-
-						break;
-					}
-
-					/**
-					 * check if we saw a \n(\r)?\n sequence 
-					 */
-					if (last_eol > 0 && 
-					    ((i - last_eol == 1) || 
-					     (i - last_eol == 2 && hctx->response_header->ptr[i - 1] == '\r'))) {
-						is_header_end = 1;
-						break;
-					}
-
-					last_eol = i;
-
-					break;
-				}
-			}
-
-			if (is_header_end) {
-				if (!is_header) {
-					/* no header, but a body */
-					if (0 != http_chunk_append_buffer(srv, con, hctx->response_header)) {
-						return FDEVENT_HANDLED_ERROR;
-					}
-				} else {
-					const char *bstart;
-					size_t blen;
-
-					/* the body starts after the EOL */
-					bstart = hctx->response_header->ptr + i;
-					blen = header_len - i;
-
-					/**
-					 * i still points to the char after the terminating EOL EOL
-					 *
-					 * put it on the last \n again
-					 */
-					i--;
-
-					/* string the last \r?\n */
-					if (i > 0 && (hctx->response_header->ptr[i - 1] == '\r')) {
-						i--;
-					}
-
-					buffer_string_set_length(hctx->response_header, i);
-
-					/* parse the response header */
-					cgi_response_parse(srv, con, p, hctx->response_header);
-
-					if (con->http_status >= 300 && con->http_status < 400) {
-						/*(con->parsed_response & HTTP_LOCATION)*/
-						size_t ulen = buffer_string_length(con->uri.path);
-						data_string *ds;
-						if (NULL != (ds = (data_string *) array_get_element(con->response.headers, "Location"))
-						    && ds->value->ptr[0] == '/'
-						    && (0 != strncmp(ds->value->ptr, con->uri.path->ptr, ulen)
-							|| (ds->value->ptr[ulen] != '\0' && ds->value->ptr[ulen] != '/' && ds->value->ptr[ulen] != '?'))
-						    && NULL == array_get_element(con->response.headers, "Set-Cookie")) {
-							if (++con->loops_per_request > 5) {
-								log_error_write(srv, __FILE__, __LINE__, "sb", "too many internal loops while processing request:", con->request.orig_uri);
-								con->http_status = 500; /* Internal Server Error */
-								con->mode = DIRECT;
-								return FDEVENT_HANDLED_FINISHED;
-							}
-
-							buffer_copy_buffer(con->request.uri, ds->value);
-
-							if (con->request.content_length) {
-								if (con->request.content_length != con->request_content_queue->bytes_in) {
-									con->keep_alive = 0;
-								}
-								con->request.content_length = 0;
-								chunkqueue_reset(con->request_content_queue);
-							}
-
-							if (con->http_status != 307 && con->http_status != 308) {
-								/* Note: request body (if any) sent to initial dynamic handler
-								 * and is not available to the internal redirect */
-								con->request.http_method = HTTP_METHOD_GET;
-							}
-
-							connection_response_reset(srv, con); /*(includes con->http_status = 0)*/
-
-							con->mode = DIRECT;
-							return FDEVENT_HANDLED_COMEBACK;
-						}
-					}
-
-					if (hctx->conf.xsendfile_allow) {
-						data_string *ds;
-						if (NULL != (ds = (data_string *) array_get_element(con->response.headers, "X-Sendfile"))) {
-							http_response_xsendfile(srv, con, ds->value, hctx->conf.xsendfile_docroot);
-							return FDEVENT_HANDLED_FINISHED;
-						}
-					}
-
-					if (blen > 0) {
-						if (0 != http_chunk_append_mem(srv, con, bstart, blen)) {
-							return FDEVENT_HANDLED_ERROR;
-						}
-					}
-				}
-
-				con->file_started = 1;
-			} else {
-				/*(reuse MAX_HTTP_REQUEST_HEADER as max size for response headers from backends)*/
-				if (header_len > MAX_HTTP_REQUEST_HEADER) {
-					log_error_write(srv, __FILE__, __LINE__, "sb", "response headers too large for", con->uri.path);
-					con->http_status = 502; /* Bad Gateway */
-					con->mode = DIRECT;
-					return FDEVENT_HANDLED_FINISHED;
-				}
-			}
-		} else {
-			if (0 != http_chunk_append_buffer(srv, con, hctx->response)) {
-				return FDEVENT_HANDLED_ERROR;
-			}
-			if ((con->conf.stream_response_body & FDEVENT_STREAM_RESPONSE_BUFMIN)
-			    && chunkqueue_length(con->write_queue) > 65536 - 4096) {
-				if (!con->is_writable) {
-					/*(defer removal of FDEVENT_IN interest since
-					 * connection_state_machine() might be able to send data
-					 * immediately, unless !con->is_writable, where
-					 * connection_state_machine() might not loop back to call
-					 * mod_cgi_handle_subrequest())*/
-					fdevent_event_clr(srv->ev, &(hctx->fde_ndx), hctx->fd, FDEVENT_IN);
-				}
-				break;
-			}
-		}
-
-#if 0
-		log_error_write(srv, __FILE__, __LINE__, "ddss", con->fd, hctx->fd, connection_get_state(con->state), b->ptr);
-#endif
-	}
-
-	return FDEVENT_HANDLED_NOT_FINISHED;
-}
 
 static void cgi_connection_close_fdtocgi(server *srv, handler_ctx *hctx) {
 	/*(closes only hctx->fdtocgi)*/
 	fdevent_event_del(srv->ev, &(hctx->fde_ndx_tocgi), hctx->fdtocgi);
-	fdevent_unregister(srv->ev, hctx->fdtocgi);
+	/*fdevent_unregister(srv->ev, hctx->fdtocgi);*//*(handled below)*/
 	fdevent_sched_close(srv->ev, hctx->fdtocgi, 0);
 	hctx->fdtocgi = -1;
 }
 
 static void cgi_connection_close(server *srv, handler_ctx *hctx) {
-	int status;
-	pid_t pid;
 	plugin_data *p = hctx->plugin_data;
 	connection *con = hctx->remote_conn;
 
-#ifndef __WIN32
-
 	/* the connection to the browser went away, but we still have a connection
 	 * to the CGI script
 	 *
@@ -642,7 +278,7 @@ static void cgi_connection_close(server *srv, handler_ctx *hctx) {
 	if (hctx->fd != -1) {
 		/* close connection to the cgi-script */
 		fdevent_event_del(srv->ev, &(hctx->fde_ndx), hctx->fd);
-		fdevent_unregister(srv->ev, hctx->fd);
+		/*fdevent_unregister(srv->ev, hctx->fd);*//*(handled below)*/
 		fdevent_sched_close(srv->ev, hctx->fd, 0);
 	}
 
@@ -650,64 +286,14 @@ static void cgi_connection_close(server *srv, handler_ctx *hctx) {
 		cgi_connection_close_fdtocgi(srv, hctx); /*(closes only hctx->fdtocgi)*/
 	}
 
-	pid = hctx->pid;
+	if (hctx->pid > 0) {
+		cgi_pid_kill(p, hctx->pid);
+	}
 
 	con->plugin_ctx[p->id] = NULL;
 
 	cgi_handler_ctx_free(hctx);
 
-	/* if waitpid hasn't been called by response.c yet, do it here */
-	if (pid) {
-		/* check if the CGI-script is already gone */
-		switch(waitpid(pid, &status, WNOHANG)) {
-		case 0:
-			/* not finished yet */
-#if 0
-			log_error_write(srv, __FILE__, __LINE__, "sd", "(debug) child isn't done yet, pid:", pid);
-#endif
-			break;
-		case -1:
-			/* */
-			if (errno == EINTR) break;
-
-			/*
-			 * errno == ECHILD happens if _subrequest catches the process-status before
-			 * we have read the response of the cgi process
-			 *
-			 * -> catch status
-			 * -> WAIT_FOR_EVENT
-			 * -> read response
-			 * -> we get here with waitpid == ECHILD
-			 *
-			 */
-			if (errno != ECHILD) {
-				log_error_write(srv, __FILE__, __LINE__, "ss", "waitpid failed: ", strerror(errno));
-			}
-			/* anyway: don't wait for it anymore */
-			pid = 0;
-			break;
-		default:
-			if (WIFEXITED(status)) {
-#if 0
-				log_error_write(srv, __FILE__, __LINE__, "sd", "(debug) cgi exited fine, pid:", pid);
-#endif
-			} else {
-				log_error_write(srv, __FILE__, __LINE__, "sd", "cgi died, pid:", pid);
-			}
-
-			pid = 0;
-			break;
-		}
-
-		if (pid) {
-			kill(pid, SIGTERM);
-
-			/* cgi-script is still alive, queue the PID for removal */
-			cgi_pid_add(srv, p, pid);
-		}
-	}
-#endif
-
 	/* finish response (if not already con->file_started, con->file_finished) */
 	if (con->mode == p->id) {
 		http_response_backend_done(srv, con);
@@ -765,31 +351,57 @@ static handler_t cgi_handle_fdevent_send (server *srv, void *ctx, int revents) {
 }
 
 
+static handler_t cgi_response_headers(server *srv, connection *con, struct http_response_opts_t *opts) {
+    /* response headers just completed */
+    handler_ctx *hctx = (handler_ctx *)opts->pdata;
+
+    if (con->parsed_response & HTTP_UPGRADE) {
+        if (hctx->conf.upgrade && con->http_status == 101) {
+            /* 101 Switching Protocols; transition to transparent proxy */
+            http_response_upgrade_read_body_unknown(srv, con);
+        }
+        else {
+            con->parsed_response &= ~HTTP_UPGRADE;
+          #if 0
+            /* preserve prior questionable behavior; likely broken behavior
+             * anyway if backend thinks connection is being upgraded but client
+             * does not receive Connection: upgrade */
+            response_header_overwrite(srv, con, CONST_STR_LEN("Upgrade"),
+                                                CONST_STR_LEN(""));
+          #endif
+        }
+    }
+
+    if (hctx->conf.upgrade && !(con->parsed_response & HTTP_UPGRADE)) {
+        chunkqueue *cq = con->request_content_queue;
+        hctx->conf.upgrade = 0;
+        if (cq->bytes_out == (off_t)con->request.content_length) {
+            cgi_connection_close_fdtocgi(srv, hctx); /*(closes hctx->fdtocgi)*/
+        }
+    }
+
+    return HANDLER_GO_ON;
+}
+
+
 static int cgi_recv_response(server *srv, handler_ctx *hctx) {
-		switch (cgi_demux_response(srv, hctx)) {
-		case FDEVENT_HANDLED_NOT_FINISHED:
-			break;
-		case FDEVENT_HANDLED_FINISHED:
-			/* we are done */
-
-#if 0
-			log_error_write(srv, __FILE__, __LINE__, "ddss", con->fd, hctx->fd, connection_get_state(con->state), "finished");
-#endif
+		switch (http_response_read(srv, hctx->remote_conn, &hctx->opts,
+					   hctx->response, hctx->fd, &hctx->fde_ndx)) {
+		default:
+			return HANDLER_GO_ON;
+		case HANDLER_ERROR:
+			http_response_backend_error(srv, hctx->remote_conn);
+			/* fall through */
+		case HANDLER_FINISHED:
 			cgi_connection_close(srv, hctx);
-
-			/* if we get a IN|HUP and have read everything don't exec the close twice */
 			return HANDLER_FINISHED;
-		case FDEVENT_HANDLED_COMEBACK:
-			cgi_connection_close(srv, hctx);
+		case HANDLER_COMEBACK:
+			/* hctx->conf.local_redir */
+			connection_response_reset(srv, hctx->remote_conn); /*(includes con->http_status = 0)*/
+			plugins_call_connection_reset(srv, hctx->remote_conn);
+			/*cgi_connection_close(srv, hctx);*//*(already cleaned up and hctx is now invalid)*/
 			return HANDLER_COMEBACK;
-		case FDEVENT_HANDLED_ERROR:
-			log_error_write(srv, __FILE__, __LINE__, "s", "demuxer failed: ");
-
-			cgi_connection_close(srv, hctx);
-			return HANDLER_FINISHED;
 		}
-
-		return HANDLER_GO_ON;
 }
 
 
@@ -817,25 +429,19 @@ static handler_t cgi_handle_fdevent(server *srv, void *ctx, int revents) {
 				rc = cgi_recv_response(srv,hctx);/*(might invalidate hctx)*/
 			} while (rc == HANDLER_GO_ON);           /*(unless HANDLER_GO_ON)*/
 			return rc; /* HANDLER_FINISHED or HANDLER_COMEBACK or HANDLER_ERROR */
-		} else if (!buffer_string_is_empty(hctx->response_header)) {
+		} else if (!buffer_string_is_empty(hctx->response)) {
 			/* unfinished header package which is a body in reality */
 			con->file_started = 1;
-			if (0 != http_chunk_append_buffer(srv, con, hctx->response_header)) {
+			if (0 != http_chunk_append_buffer(srv, con, hctx->response)) {
 				cgi_connection_close(srv, hctx);
 				return HANDLER_ERROR;
 			}
-		} else {
-# if 0
-			log_error_write(srv, __FILE__, __LINE__, "sddd", "got HUP from cgi", con->fd, hctx->fd, revents);
-# endif
+			if (0 == con->http_status) con->http_status = 200; /* OK */
 		}
 		cgi_connection_close(srv, hctx);
 	} else if (revents & FDEVENT_ERR) {
 		/* kill all connections to the cgi process */
 		cgi_connection_close(srv, hctx);
-#if 1
-		log_error_write(srv, __FILE__, __LINE__, "s", "cgi-FDEVENT_ERR");
-#endif
 		return HANDLER_ERROR;
 	}
 
@@ -894,7 +500,7 @@ static ssize_t cgi_write_file_chunk_mmap(server *srv, connection *con, int fd, c
 	off_t offset, toSend, file_end;
 	ssize_t r;
 	size_t mmap_offset, mmap_avail;
-	char *data;
+	char *data = NULL;
 
 	force_assert(NULL != c);
 	force_assert(FILE_CHUNK == c->type);
@@ -909,7 +515,7 @@ static ssize_t cgi_write_file_chunk_mmap(server *srv, connection *con, int fd, c
 		return 0;
 	}
 
-	/*(simplified from network_write_no_mmap.c:network_open_file_chunk())*/
+	/*(simplified from chunk.c:chunkqueue_open_file_chunk())*/
 	UNUSED(con);
 	if (-1 == c->file.fd) {
 		if (-1 == (c->file.fd = fdevent_open_cloexec(c->file.name->ptr, O_RDONLY, 0))) {
@@ -1046,7 +652,7 @@ static int cgi_write_request(server *srv, handler_ctx *hctx, int fd) {
 		}
 	}
 
-	if (cq->bytes_out == (off_t)con->request.content_length) {
+	if (cq->bytes_out == (off_t)con->request.content_length && !hctx->conf.upgrade) {
 		/* sent all request body input */
 		/* close connection to the cgi-script */
 		if (-1 == hctx->fdtocgi) { /*(received request body sent in initial send to pipe buffer)*/
@@ -1085,17 +691,16 @@ static int cgi_write_request(server *srv, handler_ctx *hctx, int fd) {
 }
 
 static int cgi_create_env(server *srv, connection *con, plugin_data *p, handler_ctx *hctx, buffer *cgi_handler) {
-	pid_t pid;
-
+	char_array env;
+	char *args[3];
 	int to_cgi_fds[2];
 	int from_cgi_fds[2];
-	struct stat st;
+	int dfd = -1;
 	UNUSED(p);
 
-#ifndef __WIN32
-
 	if (!buffer_string_is_empty(cgi_handler)) {
 		/* stat the exec file */
+		struct stat st;
 		if (-1 == (stat(cgi_handler->ptr, &st))) {
 			log_error_write(srv, __FILE__, __LINE__, "sbss",
 					"stat for cgi-handler", cgi_handler,
@@ -1108,42 +713,20 @@ static int cgi_create_env(server *srv, connection *con, plugin_data *p, handler_
 		log_error_write(srv, __FILE__, __LINE__, "ss", "pipe failed:", strerror(errno));
 		return -1;
 	}
-
 	if (pipe_cloexec(from_cgi_fds)) {
 		close(to_cgi_fds[0]);
 		close(to_cgi_fds[1]);
 		log_error_write(srv, __FILE__, __LINE__, "ss", "pipe failed:", strerror(errno));
 		return -1;
 	}
+	fdevent_setfd_cloexec(to_cgi_fds[1]);
+	fdevent_setfd_cloexec(from_cgi_fds[0]);
 
-	/* fork, execve */
-	switch (pid = fork()) {
-	case 0: {
-		/* child */
-		char **args;
-		int argc;
+	{
 		int i = 0;
-		char_array env;
-		char *c;
 		const char *s;
 		http_cgi_opts opts = { 0, 0, NULL, NULL };
 
-		/* move stdout to from_cgi_fd[1] */
-		dup2(from_cgi_fds[1], STDOUT_FILENO);
-	      #ifndef FD_CLOEXEC
-		close(from_cgi_fds[1]);
-		/* not needed */
-		close(from_cgi_fds[0]);
-	      #endif
-
-		/* move the stdin to to_cgi_fd[0] */
-		dup2(to_cgi_fds[0], STDIN_FILENO);
-	      #ifndef FD_CLOEXEC
-		close(to_cgi_fds[0]);
-		/* not needed */
-		close(to_cgi_fds[1]);
-	      #endif
-
 		/* create environment */
 		env.ptr = NULL;
 		env.size = 0;
@@ -1174,9 +757,6 @@ static int cgi_create_env(server *srv, connection *con, plugin_data *p, handler_
 		env.ptr[env.used] = NULL;
 
 		/* set up args */
-		argc = 3;
-		args = malloc(sizeof(*args) * argc);
-		force_assert(args);
 		i = 0;
 
 		if (!buffer_string_is_empty(cgi_handler)) {
@@ -1184,56 +764,39 @@ static int cgi_create_env(server *srv, connection *con, plugin_data *p, handler_
 		}
 		args[i++] = con->physical.path->ptr;
 		args[i  ] = NULL;
-
-		/* search for the last / */
-		if (NULL != (c = strrchr(con->physical.path->ptr, '/'))) {
-			/* handle special case of file in root directory */
-			const char* physdir = (c == con->physical.path->ptr) ? "/" : con->physical.path->ptr;
-
-			/* temporarily shorten con->physical.path to directory without terminating '/' */
-			*c = '\0';
-			/* change to the physical directory */
-			if (-1 == chdir(physdir)) {
-				log_error_write(srv, __FILE__, __LINE__, "ssb", "chdir failed:", strerror(errno), con->physical.path);
-			}
-			*c = '/';
-		}
-
-		/* we don't need the client socket */
-		for (i = 3; i < 256; i++) {
-			if (i != srv->errorlog_fd) close(i);
-		}
-
-		/* exec the cgi */
-		execve(args[0], args, env.ptr);
-
-		/* most log files may have been closed/redirected by this point,
-		 * though stderr might still point to lighttpd.breakage.log */
-		perror(args[0]);
-		_exit(1);
 	}
-	case -1:
-		/* error */
+
+	dfd = fdevent_open_dirname(con->physical.path->ptr);
+	if (-1 == dfd) {
+		log_error_write(srv, __FILE__, __LINE__, "ssb", "open dirname failed:", strerror(errno), con->physical.path);
+	}
+
+	hctx->pid = (dfd >= 0) ? fdevent_fork_execve(args[0], args, env.ptr, to_cgi_fds[0], from_cgi_fds[1], -1, dfd) : -1;
+
+	for (size_t i = 0; i < env.used; ++i) free(env.ptr[i]);
+	free(env.ptr);
+
+	if (-1 == hctx->pid) {
+		/* log error with errno prior to calling close() (might change errno) */
 		log_error_write(srv, __FILE__, __LINE__, "ss", "fork failed:", strerror(errno));
+		if (-1 != dfd) close(dfd);
 		close(from_cgi_fds[0]);
 		close(from_cgi_fds[1]);
 		close(to_cgi_fds[0]);
 		close(to_cgi_fds[1]);
 		return -1;
-	default: {
-		/* parent process */
-
+	} else {
+		if (-1 != dfd) close(dfd);
 		close(from_cgi_fds[1]);
 		close(to_cgi_fds[0]);
 
-		/* register PID and wait for them asynchronously */
-
-		hctx->pid = pid;
 		hctx->fd = from_cgi_fds[0];
 		hctx->fde_ndx = -1;
 
 		++srv->cur_fds;
 
+		cgi_pid_add(p, hctx->pid, hctx);
+
 		if (0 == con->request.content_length) {
 			close(to_cgi_fds[1]);
 		} else {
@@ -1262,14 +825,8 @@ static int cgi_create_env(server *srv, connection *con, plugin_data *p, handler_
 		}
 		fdevent_event_set(srv->ev, &(hctx->fde_ndx), hctx->fd, FDEVENT_IN);
 
-		break;
+		return 0;
 	}
-	}
-
-	return 0;
-#else
-	return -1;
-#endif
 }
 
 static buffer * cgi_get_handler(array *a, buffer *fn) {
@@ -1297,6 +854,8 @@ static int mod_cgi_patch_connection(server *srv, connection *con, plugin_data *p
 
 	PATCH(cgi);
 	PATCH(execute_x_only);
+	PATCH(local_redir);
+	PATCH(upgrade);
 	PATCH(xsendfile_allow);
 	PATCH(xsendfile_docroot);
 
@@ -1316,6 +875,10 @@ static int mod_cgi_patch_connection(server *srv, connection *con, plugin_data *p
 				PATCH(cgi);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("cgi.execute-x-only"))) {
 				PATCH(execute_x_only);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("cgi.local-redir"))) {
+				PATCH(local_redir);
+			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("cgi.upgrade"))) {
+				PATCH(upgrade);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("cgi.x-sendfile"))) {
 				PATCH(xsendfile_allow);
 			} else if (buffer_is_equal_string(du->key, CONST_STR_LEN("cgi.x-sendfile-docroot"))) {
@@ -1360,6 +923,18 @@ URIHANDLER_FUNC(cgi_is_handled) {
 		hctx->plugin_data = p;
 		hctx->cgi_handler = cgi_handler;
 		memcpy(&hctx->conf, &p->conf, sizeof(plugin_config));
+		hctx->conf.upgrade =
+		  hctx->conf.upgrade
+		  && con->request.http_version == HTTP_VERSION_1_1
+		  && NULL != array_get_element_klen(con->request.headers, CONST_STR_LEN("Upgrade"));
+		hctx->opts.fdfmt = S_IFIFO;
+		hctx->opts.backend = BACKEND_CGI;
+		hctx->opts.authorizer = 0;
+		hctx->opts.local_redir = hctx->conf.local_redir;
+		hctx->opts.xsendfile_allow = hctx->conf.xsendfile_allow;
+		hctx->opts.xsendfile_docroot = hctx->conf.xsendfile_docroot;
+		hctx->opts.pdata = hctx;
+		hctx->opts.headers = cgi_response_headers;
 		con->plugin_ctx[p->id] = hctx;
 		con->mode = p->id;
 	}
@@ -1367,63 +942,6 @@ URIHANDLER_FUNC(cgi_is_handled) {
 	return HANDLER_GO_ON;
 }
 
-TRIGGER_FUNC(cgi_trigger) {
-	plugin_data *p = p_d;
-	size_t ndx;
-	/* the trigger handle only cares about lonely PID which we have to wait for */
-#ifndef __WIN32
-
-	for (ndx = 0; ndx < p->cgi_pid.used; ndx++) {
-		int status;
-
-		switch(waitpid(p->cgi_pid.ptr[ndx], &status, WNOHANG)) {
-		case 0:
-			/* not finished yet */
-#if 0
-			log_error_write(srv, __FILE__, __LINE__, "sd", "(debug) child isn't done yet, pid:", p->cgi_pid.ptr[ndx]);
-#endif
-			break;
-		case -1:
-			if (errno == ECHILD) {
-				/* someone else called waitpid... remove the pid to stop looping the error each time */
-				log_error_write(srv, __FILE__, __LINE__, "s", "cgi child vanished, probably someone else called waitpid");
-
-				cgi_pid_del(srv, p, p->cgi_pid.ptr[ndx]);
-				ndx--;
-				continue;
-			}
-
-			log_error_write(srv, __FILE__, __LINE__, "ss", "waitpid failed: ", strerror(errno));
-
-			return HANDLER_ERROR;
-		default:
-
-			if (WIFEXITED(status)) {
-#if 0
-				log_error_write(srv, __FILE__, __LINE__, "sd", "(debug) cgi exited fine, pid:", p->cgi_pid.ptr[ndx]);
-#endif
-			} else if (WIFSIGNALED(status)) {
-				/* FIXME: what if we killed the CGI script with a kill(..., SIGTERM) ?
-				 */
-				if (WTERMSIG(status) != SIGTERM) {
-					log_error_write(srv, __FILE__, __LINE__, "sd", "cleaning up CGI: process died with signal", WTERMSIG(status));
-				}
-			} else {
-				log_error_write(srv, __FILE__, __LINE__, "s", "cleaning up CGI: ended unexpectedly");
-			}
-
-			cgi_pid_del(srv, p, p->cgi_pid.ptr[ndx]);
-			/* del modified the buffer structure
-			 * and copies the last entry to the current one
-			 * -> recheck the current index
-			 */
-			ndx--;
-		}
-	}
-#endif
-	return HANDLER_GO_ON;
-}
-
 /*
  * - HANDLER_GO_ON : not our job
  * - HANDLER_FINISHED: got response
@@ -1442,9 +960,10 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 		if (chunkqueue_length(con->write_queue) > 65536 - 4096) {
 			fdevent_event_clr(srv->ev, &(hctx->fde_ndx), hctx->fd, FDEVENT_IN);
 		} else if (!(fdevent_event_get_interest(srv->ev, hctx->fd) & FDEVENT_IN)) {
-			/* optimistic read from backend, which might re-enable FDEVENT_IN */
+			/* optimistic read from backend */
 			handler_t rc = cgi_recv_response(srv, hctx); /*(might invalidate hctx)*/
 			if (rc != HANDLER_GO_ON) return rc;          /*(unless HANDLER_GO_ON)*/
+			fdevent_event_add(srv->ev, &(hctx->fde_ndx), hctx->fd, FDEVENT_IN);
 		}
 	}
 
@@ -1481,9 +1000,6 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 
 			return HANDLER_FINISHED;
 		}
-#if 0
-	log_error_write(srv, __FILE__, __LINE__, "sdd", "subrequest, pid =", hctx, hctx->pid);
-#endif
 	} else if (!chunkqueue_is_empty(con->request_content_queue)) {
 		if (0 != cgi_write_request(srv, hctx, hctx->fdtocgi)) {
 			cgi_connection_close(srv, hctx);
@@ -1496,6 +1012,38 @@ SUBREQUEST_FUNC(mod_cgi_handle_subrequest) {
 }
 
 
+static handler_t cgi_waitpid_cb(server *srv, void *p_d, pid_t pid, int status) {
+    plugin_data *p = (plugin_data *)p_d;
+    for (size_t i = 0; i < p->cgi_pid.used; ++i) {
+        handler_ctx *hctx;
+        if (pid != p->cgi_pid.ptr[i].pid) continue;
+
+        hctx = (handler_ctx *)p->cgi_pid.ptr[i].ctx;
+        if (hctx) hctx->pid = -1;
+        cgi_pid_del(p, i);
+
+        if (WIFEXITED(status)) {
+            /* (skip logging (non-zero) CGI exit; might be very noisy) */
+        }
+        else if (WIFSIGNALED(status)) {
+            /* ignore SIGTERM if sent by cgi_connection_close() (NULL == hctx)*/
+            if (WTERMSIG(status) != SIGTERM || NULL != hctx) {
+                log_error_write(srv, __FILE__, __LINE__, "sdsd", "CGI pid", pid,
+                                "died with signal", WTERMSIG(status));
+            }
+        }
+        else {
+            log_error_write(srv, __FILE__, __LINE__, "sds",
+                            "CGI pid", pid, "ended unexpectedly");
+        }
+
+        return HANDLER_FINISHED;
+    }
+
+    return HANDLER_GO_ON;
+}
+
+
 int mod_cgi_plugin_init(plugin *p);
 int mod_cgi_plugin_init(plugin *p) {
 	p->version     = LIGHTTPD_VERSION_ID;
@@ -1504,7 +1052,7 @@ int mod_cgi_plugin_init(plugin *p) {
 	p->connection_reset = cgi_connection_close_callback;
 	p->handle_subrequest_start = cgi_is_handled;
 	p->handle_subrequest = mod_cgi_handle_subrequest;
-	p->handle_trigger = cgi_trigger;
+	p->handle_waitpid = cgi_waitpid_cb;
 	p->init           = mod_cgi_init;
 	p->cleanup        = mod_cgi_free;
 	p->set_defaults   = mod_fastcgi_set_defaults;
