@@ -14,8 +14,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
-#include <unistd.h>
-#include <stdio.h>
 
 /* init the plugin data */
 INIT_FUNC(mod_cml_init) {
@@ -91,7 +89,7 @@ SETDEFAULTS_FUNC(mod_cml_set_defaults) {
 		data_config const* config = (data_config const*)srv->config_context->data[i];
 		plugin_config *s;
 
-		s = malloc(sizeof(plugin_config));
+		s = calloc(1, sizeof(plugin_config));
 		s->ext    = buffer_init();
 		s->mc_hosts       = array_init();
 		s->mc_namespace   = buffer_init();
@@ -111,6 +109,12 @@ SETDEFAULTS_FUNC(mod_cml_set_defaults) {
 			return HANDLER_ERROR;
 		}
 
+		if (!array_is_vlist(s->mc_hosts)) {
+			log_error_write(srv, __FILE__, __LINE__, "s",
+					"unexpected value for cml.memcache-hosts; expected list of \"host\"");
+			return HANDLER_ERROR;
+		}
+
 		if (s->mc_hosts->used) {
 #if defined(USE_MEMCACHED)
 			buffer *option_string = buffer_init();
