@@ -5,11 +5,9 @@
 #include "buffer.h"
 #include "response.h"
 #include "http_chunk.h"
-#include "stat_cache.h"
 
 #include "plugin.h"
 
-#include <ctype.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -102,6 +100,12 @@ SETDEFAULTS_FUNC(mod_flv_streaming_set_defaults) {
 		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
 			return HANDLER_ERROR;
 		}
+
+		if (!array_is_vlist(s->extensions)) {
+			log_error_write(srv, __FILE__, __LINE__, "s",
+					"unexpected value for flv-streaming.extensions; expected list of \"ext\"");
+			return HANDLER_ERROR;
+		}
 	}
 
 	return HANDLER_GO_ON;
@@ -210,33 +214,39 @@ URIHANDLER_FUNC(mod_flv_streaming_path_handler) {
 
 		if (0 == strncmp(con->physical.path->ptr + s_len - ct_len, ds->value->ptr, ct_len)) {
 			data_string *get_param;
-			int start;
+			off_t start = 0, len = -1;
 			char *err = NULL;
 			/* if there is a start=[0-9]+ in the header use it as start,
-			 * otherwise send the full file */
+			 * otherwise set start to beginning of file */
+			/* if there is a end=[0-9]+ in the header use it as end pos,
+			 * otherwise send rest of file, starting from start */
 
 			array_reset(p->get_params);
 			buffer_copy_buffer(p->query_str, con->uri.query);
 			split_get_params(p->get_params, p->query_str);
 
-			if (NULL == (get_param = (data_string *)array_get_element(p->get_params, "start"))) {
-				return HANDLER_GO_ON;
+			if (NULL != (get_param = (data_string *)array_get_element(p->get_params, "start"))) {
+				if (buffer_string_is_empty(get_param->value)) return HANDLER_GO_ON;
+				start = strtoll(get_param->value->ptr, &err, 10);
+				if (*err != '\0') return HANDLER_GO_ON;
+				if (start < 0) return HANDLER_GO_ON;
 			}
 
-			/* too short */
-			if (buffer_string_is_empty(get_param->value)) return HANDLER_GO_ON;
-
-			/* check if it is a number */
-			start = strtol(get_param->value->ptr, &err, 10);
-			if (*err != '\0') {
+			if (NULL != (get_param = (data_string *)array_get_element(p->get_params, "end"))) {
+				off_t end;
+				if (buffer_string_is_empty(get_param->value)) return HANDLER_GO_ON;
+				end = strtoll(get_param->value->ptr, &err, 10);
+				if (*err != '\0') return HANDLER_GO_ON;
+				if (end < 0) return HANDLER_GO_ON;
+				len = (start < end ? end - start : start - end) + 1;
+			}
+			else if (0 == start) {
 				return HANDLER_GO_ON;
 			}
 
-			if (start <= 0) return HANDLER_GO_ON;
-
 			/* let's build a flv header */
 			http_chunk_append_mem(srv, con, CONST_STR_LEN("FLV\x1\x1\0\0\0\x9\0\0\0\x9"));
-			if (0 != http_chunk_append_file_range(srv, con, con->physical.path, start, -1)) {
+			if (0 != http_chunk_append_file_range(srv, con, con->physical.path, start, len)) {
 				chunkqueue_reset(con->write_queue);
 				return HANDLER_GO_ON;
 			}
