@@ -8,7 +8,6 @@
 
 #include "stat_cache.h"
 
-#include <ctype.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
@@ -100,6 +99,12 @@ SETDEFAULTS_FUNC(mod_indexfile_set_defaults) {
 		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
 			return HANDLER_ERROR;
 		}
+
+		if (!array_is_vlist(s->indexfiles)) {
+			log_error_write(srv, __FILE__, __LINE__, "s",
+					"unexpected value for index-file.names; expected list of \"file\"");
+			return HANDLER_ERROR;
+		}
 	}
 
 	return HANDLER_GO_ON;
