@@ -1,6 +1,7 @@
 #include "first.h"
 
 #include "base.h"
+#include "fdevent.h"
 #include "log.h"
 #include "buffer.h"
 
@@ -10,17 +11,15 @@
 
 #include "mod_ssi.h"
 
-#include "inet_ntop_cache.h"
-
 #include "sys-socket.h"
 
 #include <sys/types.h>
+#include "sys-strings.h"
+#include <sys/wait.h>
 
 #include <ctype.h>
 #include <stdlib.h>
-#include <stdio.h>
 #include <string.h>
-#include <strings.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <time.h>
@@ -30,10 +29,6 @@
 # include <pwd.h>
 #endif
 
-#ifdef HAVE_FORK
-# include <sys/wait.h>
-#endif
-
 #ifdef HAVE_SYS_FILIO_H
 # include <sys/filio.h>
 #endif
@@ -146,6 +141,12 @@ SETDEFAULTS_FUNC(mod_ssi_set_defaults) {
 		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
 			return HANDLER_ERROR;
 		}
+
+		if (!array_is_vlist(s->ssi_extension)) {
+			log_error_write(srv, __FILE__, __LINE__, "s",
+					"unexpected value for ssi.extension; expected list of \"ext\"");
+			return HANDLER_ERROR;
+		}
 	}
 
 	return HANDLER_GO_ON;
@@ -462,8 +463,8 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 			data_string *ds;
 			/* check if it is a cgi-var or a ssi-var */
 
-			if (NULL != (ds = (data_string *)array_get_element(p->ssi_cgi_env, var_val)) ||
-			    NULL != (ds = (data_string *)array_get_element(p->ssi_vars, var_val))) {
+			if (NULL != (ds = (data_string *)array_get_element_klen(p->ssi_cgi_env, var_val, strlen(var_val))) ||
+			    NULL != (ds = (data_string *)array_get_element_klen(p->ssi_vars, var_val, strlen(var_val)))) {
 				chunkqueue_append_mem(con->write_queue, CONST_BUF_LEN(ds->value));
 			} else {
 				chunkqueue_append_mem(con->write_queue, CONST_STR_LEN("(none)"));
@@ -759,7 +760,8 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 	case SSI_EXEC: {
 		const char *cmd = NULL;
 		pid_t pid;
-		int from_exec_fds[2];
+		chunk *c;
+		char *args[4];
 
 		if (!p->conf.ssi_exec) { /* <!--#exec ... --> disabled by config */
 			break;
@@ -777,105 +779,48 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 
 		if (p->if_is_false) break;
 
-		/* create a return pipe and send output to the html-page
-		 *
+		/*
 		 * as exec is assumed evil it is implemented synchronously
 		 */
 
 		if (!cmd) break;
-#ifdef HAVE_FORK
-		if (pipe(from_exec_fds)) {
-			log_error_write(srv, __FILE__, __LINE__, "ss",
-					"pipe failed: ", strerror(errno));
-			return -1;
-		}
 
-		/* fork, execve */
-		switch (pid = fork()) {
-		case 0: {
-			/* move stdout to from_rrdtool_fd[1] */
-			close(STDOUT_FILENO);
-			dup2(from_exec_fds[1], STDOUT_FILENO);
-			close(from_exec_fds[1]);
-			/* not needed */
-			close(from_exec_fds[0]);
+		/* send cmd output to a temporary file */
+		if (0 != chunkqueue_append_mem_to_tempfile(srv, con->write_queue, "", 0)) break;
+		c = con->write_queue->last;
 
-			/* close stdin */
-			close(STDIN_FILENO);
+		*(const char **)&args[0] = "/bin/sh";
+		*(const char **)&args[1] = "-c";
+		*(const char **)&args[2] = cmd;
+		args[3] = NULL;
 
-			execl("/bin/sh", "sh", "-c", cmd, (char *)NULL);
-
-			log_error_write(srv, __FILE__, __LINE__, "sss", "spawing exec failed:", strerror(errno), cmd);
-
-			/* */
-			SEGFAULT();
-			break;
-		}
-		case -1:
-			/* error */
-			log_error_write(srv, __FILE__, __LINE__, "ss", "fork failed:", strerror(errno));
-			break;
-		default: {
-			/* father */
+		/*(expects STDIN_FILENO open to /dev/null)*/
+		pid = fdevent_fork_execve(args[0], args, NULL, -1, c->file.fd, -1, -1);
+		if (-1 == pid) {
+			log_error_write(srv, __FILE__, __LINE__, "sss", "spawning exec failed:", strerror(errno), cmd);
+		} else {
+			struct stat stb;
 			int status;
-			ssize_t r;
-			int was_interrupted = 0;
-
-			close(from_exec_fds[1]);
 
 			/* wait for the client to end */
+			/* NOTE: synchronous; blocks entire lighttpd server */
 
 			/*
 			 * OpenBSD and Solaris send a EINTR on SIGCHILD even if we ignore it
 			 */
-			do {
-				if (-1 == waitpid(pid, &status, 0)) {
-					if (errno == EINTR) {
-						was_interrupted++;
-					} else {
-						was_interrupted = 0;
-						log_error_write(srv, __FILE__, __LINE__, "ss", "waitpid failed:", strerror(errno));
-					}
-				} else if (WIFEXITED(status)) {
-					int toread;
-					/* read everything from client and paste it into the output */
-					was_interrupted = 0;
-
-					while(1) {
-						if (ioctl(from_exec_fds[0], FIONREAD, &toread)) {
-							log_error_write(srv, __FILE__, __LINE__, "s",
-								"unexpected end-of-file (perhaps the ssi-exec process died)");
-							return -1;
-						}
-
-						if (toread > 0) {
-							char *mem;
-							size_t mem_len;
-
-							chunkqueue_get_memory(con->write_queue, &mem, &mem_len, 0, toread);
-							r = read(from_exec_fds[0], mem, mem_len);
-							chunkqueue_use_memory(con->write_queue, r > 0 ? r : 0);
-
-							if (r < 0) break; /* read failed */
-						} else {
-							break;
-						}
-					}
-				} else {
-					was_interrupted = 0;
-					log_error_write(srv, __FILE__, __LINE__, "s", "process exited abnormally");
+			while (-1 == waitpid(pid, &status, 0)) {
+				if (errno != EINTR) {
+					log_error_write(srv, __FILE__, __LINE__, "ss", "waitpid failed:", strerror(errno));
+					break;
 				}
-			} while (was_interrupted > 0 && was_interrupted < 4); /* if waitpid() gets interrupted, retry, but max 4 times */
-
-			close(from_exec_fds[0]);
-
-			break;
+			}
+			if (!WIFEXITED(status)) {
+				log_error_write(srv, __FILE__, __LINE__, "ss", "process exited abnormally:", cmd);
+			}
+			if (0 == fstat(c->file.fd, &stb)) {
+				c->file.length = stb.st_size;
+			}
 		}
-		}
-#else
-
-		return -1;
-#endif
 
 		break;
 	}
@@ -1295,6 +1240,7 @@ static int mod_ssi_handle_request(server *srv, connection *con, handler_ctx *p)
 			st.st_mtime = include_file_last_mtime;
 
 		etag_create(con->physical.etag, &st, con->etag_flags);
+		etag_mutate(con->physical.etag, con->physical.etag);
 		response_header_overwrite(srv, con, CONST_STR_LEN("ETag"), CONST_BUF_LEN(con->physical.etag));
 
 		mtime = strftime_cache_get(srv, st.st_mtime);
