@@ -6,14 +6,11 @@
 
 #include "plugin.h"
 
-#include "stat_cache.h"
 #include "etag.h"
 #include "http_chunk.h"
 #include "response.h"
 
-#include <ctype.h>
 #include <stdlib.h>
-#include <stdio.h>
 #include <string.h>
 
 /**
@@ -110,6 +107,12 @@ SETDEFAULTS_FUNC(mod_staticfile_set_defaults) {
 		if (0 != config_insert_values_global(srv, config->value, cv, i == 0 ? T_CONFIG_SCOPE_SERVER : T_CONFIG_SCOPE_CONNECTION)) {
 			return HANDLER_ERROR;
 		}
+
+		if (!array_is_vlist(s->exclude_ext)) {
+			log_error_write(srv, __FILE__, __LINE__, "s",
+					"unexpected value for static-file.exclude-extensions; expected list of \"ext\"");
+			return HANDLER_ERROR;
+		}
 	}
 
 	return HANDLER_GO_ON;
