@@ -14,13 +14,11 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <ctype.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <stdio.h>
-#include <assert.h>
 
 #include <unistd.h>
 #include <dirent.h>
@@ -33,7 +31,8 @@
 #include <sqlite3.h>
 #endif
 
-#if defined(HAVE_LIBXML_H) && defined(HAVE_SQLITE3_H) && defined(HAVE_UUID_UUID_H)
+#if defined(HAVE_LIBXML_H) && defined(HAVE_SQLITE3_H) \
+ && defined(HAVE_UUID) && defined(HAVE_UUID_UUID_H)
 #define USE_LOCKS
 #include <uuid/uuid.h>
 #endif
@@ -935,7 +934,6 @@ static int webdav_get_live_property(server *srv, connection *con, handler_ctx *h
 	if (HANDLER_ERROR != (stat_cache_get_entry(srv, con, dst->path, &sce))) {
 		char ctime_buf[] = "2005-08-18T07:27:16Z";
 		char mtime_buf[] = "Thu, 18 Aug 2005 07:27:16 GMT";
-		size_t k;
 
 		if (0 == strcmp(prop_name, "resourcetype")) {
 			if (S_ISDIR(sce->st.st_mode)) {
@@ -949,19 +947,12 @@ static int webdav_get_live_property(server *srv, connection *con, handler_ctx *h
 				buffer_append_string_len(b, CONST_STR_LEN("<D:getcontenttype>httpd/unix-directory</D:getcontenttype>"));
 				found = 1;
 			} else if(S_ISREG(sce->st.st_mode)) {
-				for (k = 0; k < con->conf.mimetypes->used; k++) {
-					data_string *ds = (data_string *)con->conf.mimetypes->data[k];
-
-					if (buffer_is_empty(ds->key)) continue;
-
-					if (buffer_is_equal_right_len(dst->path, ds->key, buffer_string_length(ds->key))) {
-						buffer_append_string_len(b,CONST_STR_LEN("<D:getcontenttype>"));
-						buffer_append_string_buffer(b, ds->value);
-						buffer_append_string_len(b, CONST_STR_LEN("</D:getcontenttype>"));
-						found = 1;
-
-						break;
-					}
+				const buffer *type = stat_cache_mimetype_by_ext(con, CONST_BUF_LEN(dst->path));
+				if (NULL != type) {
+					buffer_append_string_len(b, CONST_STR_LEN("<D:getcontenttype>"));
+					buffer_append_string_buffer(b, type);
+					buffer_append_string_len(b, CONST_STR_LEN("</D:getcontenttype>"));
+					found = 1;
 				}
 			}
 		} else if (0 == strcmp(prop_name, "creationdate")) {
@@ -988,6 +979,7 @@ static int webdav_get_live_property(server *srv, connection *con, handler_ctx *h
 			found = 1;
 		} else if (0 == strcmp(prop_name, "getetag")) {
 			etag_create(con->physical.etag, &sce->st, con->etag_flags);
+			etag_mutate(con->physical.etag, con->physical.etag);
 			buffer_append_string_len(b, CONST_STR_LEN("<D:getetag>"));
 			buffer_append_string_buffer(b, con->physical.etag);
 			buffer_append_string_len(b, CONST_STR_LEN("</D:getetag>"));
@@ -1347,6 +1339,14 @@ SUBREQUEST_FUNC(mod_webdav_subrequest_handler_huge) {
 				con->http_status = 404;
 				return HANDLER_FINISHED;
 			}
+			else if (errno == EACCES) {
+				con->http_status = 403;
+				return HANDLER_FINISHED;
+			}
+			else {
+				con->http_status = 500;
+				return HANDLER_FINISHED;
+			}
 			break;
 		default:
 			break;
