@@ -6,11 +6,10 @@
 
 void network_accept_tcp_nagle_disable(int fd);
 
-int network_write_chunkqueue(server *srv, connection *con, chunkqueue *c, off_t max_bytes);
-
-int network_init(server *srv);
+int network_init(server *srv, int stdin_fd);
 int network_close(server *srv);
 
 int network_register_fdevents(server *srv);
+void network_unregister_sock(server *srv, server_socket *srv_socket);
 
 #endif
