@@ -32,18 +32,23 @@ typedef struct {
 	buffer *name; /* name of the plugin */
 
 	void *(* init)                       ();
+	handler_t (* priv_defaults)          (server *srv, void *p_d);
 	handler_t (* set_defaults)           (server *srv, void *p_d);
 	handler_t (* cleanup)                (server *srv, void *p_d);
 	                                                                                   /* is called ... */
 	handler_t (* handle_trigger)         (server *srv, void *p_d);                     /* once a second */
-	handler_t (* handle_sighup)          (server *srv, void *p_d);                     /* at a signup */
+	handler_t (* handle_sighup)          (server *srv, void *p_d);                     /* at a sighup */
+	handler_t (* handle_waitpid)         (server *srv, void *p_d, pid_t pid, int status); /* upon a child process exit */
 
 	handler_t (* handle_uri_raw)         (server *srv, connection *con, void *p_d);    /* after uri_raw is set */
 	handler_t (* handle_uri_clean)       (server *srv, connection *con, void *p_d);    /* after uri is set */
 	handler_t (* handle_docroot)         (server *srv, connection *con, void *p_d);    /* getting the document-root */
 	handler_t (* handle_physical)        (server *srv, connection *con, void *p_d);    /* mapping url to physical path */
+	handler_t (* handle_request_env)     (server *srv, connection *con, void *p_d);    /* (deferred env populate) */
 	handler_t (* handle_request_done)    (server *srv, connection *con, void *p_d);    /* at the end of a request */
-	handler_t (* handle_connection_close)(server *srv, connection *con, void *p_d);    /* at the end of a connection */
+	handler_t (* handle_connection_accept) (server *srv, connection *con, void *p_d);  /* after accept() socket */
+	handler_t (* handle_connection_shut_wr)(server *srv, connection *con, void *p_d);  /* done writing to socket */
+	handler_t (* handle_connection_close)  (server *srv, connection *con, void *p_d);  /* before close() of socket */
 
 
 
@@ -54,7 +59,7 @@ typedef struct {
 											    */
 	handler_t (* handle_subrequest)      (server *srv, connection *con, void *p_d);    /* */
 	handler_t (* handle_response_start)  (server *srv, connection *con, void *p_d);    /* before response headers are written */
-	handler_t (* connection_reset)       (server *srv, connection *con, void *p_d);    /* */
+	handler_t (* connection_reset)       (server *srv, connection *con, void *p_d);    /* after request done or request abort */
 	void *data;
 
 	/* dlopen handle */
@@ -69,14 +74,18 @@ handler_t plugins_call_handle_uri_clean(server *srv, connection *con);
 handler_t plugins_call_handle_subrequest_start(server *srv, connection *con);
 handler_t plugins_call_handle_subrequest(server *srv, connection *con);
 handler_t plugins_call_handle_response_start(server *srv, connection *con);
+handler_t plugins_call_handle_request_env(server *srv, connection *con);
 handler_t plugins_call_handle_request_done(server *srv, connection *con);
 handler_t plugins_call_handle_docroot(server *srv, connection *con);
 handler_t plugins_call_handle_physical(server *srv, connection *con);
+handler_t plugins_call_handle_connection_accept(server *srv, connection *con);
+handler_t plugins_call_handle_connection_shut_wr(server *srv, connection *con);
 handler_t plugins_call_handle_connection_close(server *srv, connection *con);
 handler_t plugins_call_connection_reset(server *srv, connection *con);
 
 handler_t plugins_call_handle_trigger(server *srv);
 handler_t plugins_call_handle_sighup(server *srv);
+handler_t plugins_call_handle_waitpid(server *srv, pid_t pid, int status);
 
 handler_t plugins_call_init(server *srv);
 handler_t plugins_call_set_defaults(server *srv);
