@@ -4,11 +4,14 @@
 
 #include "base.h"
 
-stat_cache *stat_cache_init(void);
-void stat_cache_free(stat_cache *fc);
+struct stat_cache;      /* declaration */
 
+int stat_cache_choose_engine (server *srv, const buffer *stat_cache_string);
+struct stat_cache *stat_cache_init(server *srv);
+void stat_cache_free(struct stat_cache *fc);
+
+const buffer * stat_cache_mimetype_by_ext(const connection *con, const char *name, size_t nlen);
 handler_t stat_cache_get_entry(server *srv, connection *con, buffer *name, stat_cache_entry **fce);
-handler_t stat_cache_handle_fdevent(server *srv, void *_fce, int revent);
 int stat_cache_open_rdonly_fstat (server *srv, connection *con, buffer *name, struct stat *st);
 
 int stat_cache_trigger_cleanup(server *srv);
