@@ -2,11 +2,11 @@
 #define _STATUS_COUNTER_H_
 #include "first.h"
 
-#include "array.h"
-#include "base.h"
-
 #include <sys/types.h>
 
+#include "base_decls.h"
+#include "array.h"
+
 data_integer *status_counter_get_counter(server *srv, const char *s, size_t len);
 int status_counter_inc(server *srv, const char *s, size_t len);
 int status_counter_dec(server *srv, const char *s, size_t len);
