@@ -10,21 +10,27 @@
 #define EINPROGRESS WSAEINPROGRESS
 #define EALREADY WSAEALREADY
 #define ECONNABORTED WSAECONNABORTED
-#define ioctl ioctlsocket
+
 #else
+
 #include <sys/socket.h>
-#include <sys/ioctl.h>
 #include <netinet/in.h>
 #include <netinet/tcp.h>
+#ifdef HAVE_SYS_UN_H
 #include <sys/un.h>
-#include <arpa/inet.h>
-
-#include <netdb.h>
-
-#ifdef HAVE_SYS_FILIO_H
-#include <sys/filio.h>  /* FIONREAD (for illumos (OpenIndiana)) */
 #endif
 
 #endif
 
+
+#ifndef INET_ADDRSTRLEN
+#define INET_ADDRSTRLEN   16
+#endif
+#ifndef INET6_ADDRSTRLEN
+#define INET6_ADDRSTRLEN  46
+#endif
+#ifndef UNIX_PATH_MAX
+#define UNIX_PATH_MAX    108
+#endif
+
 #endif
